# Note: Following stuff must be considered in a GangaRepository:
#
# * lazy loading
# * locking

from GangaCore.Core.GangaRepository import GangaRepository, RepositoryError, InaccessibleObjectError
from GangaCore.Utility.Plugin import PluginManagerError
import os
import os.path
import time
import errno
import copy
import threading

from GangaCore import GANGA_SWAN_INTEGRATION

from GangaCore.Core.GangaRepository.SessionLock import SessionLockManager, dry_run_unix_locks
from GangaCore.Core.GangaRepository.FixedLock import FixedLockManager

import GangaCore.Utility.logging

from GangaCore.Core.GangaRepository.PickleStreamer import to_file as pickle_to_file
from GangaCore.Core.GangaRepository.PickleStreamer import from_file as pickle_from_file

from GangaCore.Core.GangaRepository.VStreamer import to_file as xml_to_file
from GangaCore.Core.GangaRepository.VStreamer import from_file as xml_from_file
from GangaCore.Core.GangaRepository.VStreamer import XMLFileError

from GangaCore.GPIDev.Base.Objects import Node
from GangaCore.Core.GangaRepository.SubJobXMLList import SubJobXMLList

from GangaCore.GPIDev.Base.Proxy import isType, stripProxy, getName

from GangaCore.Utility.Config import getConfig

logger = GangaCore.Utility.logging.getLogger()

save_all_history = False

def check_app_hash(obj):
    """Writes a file safely, raises IOError on error
    Args:
        obj (GangaObject): This is an object which has a prepared application
    """

    isVerifiableApp = False
    isVerifiableAna = False

    if hasattr(obj, 'application'):
        if hasattr(obj.application, 'hash'):
            if obj.application.hash is not None:
                isVerifiableApp = True
    elif hasattr(obj, 'analysis'):
        if hasattr(obj.analysis, 'application'):
            if hasattr(obj.analysis.application, 'hash'):
                if obj.analysis.application.hash is not None:
                    isVerifiableAna = True

    if isVerifiableApp is True:
        hashable_app = stripProxy(obj.application)
    elif isVerifiableAna is True:
        hashable_app = stripProxy(obj.analysis.application)
    else:
        hashable_app = None

    if hashable_app is not None:
        if not hashable_app.calc_hash(True):
            try:
                logger.warning("%s" % hashable_app)
                logger.warning('Protected attribute(s) of %s application (associated with %s #%s) changed!' % (getName(hashable_app), getName(obj), obj._registry_id))
            except AttributeError as err:
                logger.warning('Protected attribute(s) of %s application (associated with %s) changed!!!!' % (getName(hashable_app), getName(obj)))
                logger.warning("%s" % err)
            jobObj = stripProxy(hashable_app).getJobObject()
            if jobObj is not None:
                logger.warning('Job: %s is now possibly corrupt!' % jobObj.getFQID('.'))
            logger.warning('If you knowingly circumvented the protection, ignore this message (and, optionally,')
            logger.warning('re-prepare() the application). Otherwise, please file a bug report at:')
            logger.warning('https://github.com/ganga-devs/ganga/issues/')

def safe_save(fn, _obj, to_file, ignore_subs=''):
    """Try to save the XML for this object in as safe a way as possible
    Args:
        fn (str): This is the name of the file we are to save the object to
        _obj (GangaObject): This is the object which we want to save to the file
        to_file (str): This is the method we want to use to save the to the file
        ignore_subs (str): This is the name(s) of the attribute of _obj we want to ignore in writing to disk
    """

    # Add a global lock to make absolutely sure we don't have multiple threads writing files
    # See Github Issue 185
    with safe_save.lock:

        obj = stripProxy(_obj)
        check_app_hash(obj)

        # Create the dirs
        dirname = os.path.dirname(fn)
        if not os.path.exists(dirname):
            os.makedirs(dirname)

        # Prepare new data file
        new_name = fn + '.new'
        with open(new_name, "w") as tmpfile:
            to_file(obj, tmpfile, ignore_subs)

        # everything ready so create new data file and backup old one
        if os.path.exists(new_name):

            # Do we have an old one to backup?
            if os.path.exists(fn):
                os.rename(fn, fn + "~")

            os.rename(new_name, fn)

# Global lock for above function - See issue #185
safe_save.lock = threading.Lock()

def rmrf(name, count=0):
    """
    Safely recursively remove a file/folder from disk by first moving it then removing it
    calls self and will only attempt to move/remove a file 3 times before giving up
    Args:
        count (int): This function calls itself recursively 3 times then gives up, this increments on each call
    """

    if count != 0:
        logger.debug("Trying again to remove: %s" % name)
        if count == 3:
            logger.error("Tried 3 times to remove file/folder: %s" % name)
            from GangaCore.Core.exceptions import GangaException
            raise GangaException("Failed to remove file/folder: %s" % name)

    if os.path.isdir(name):

        try:
            remove_name = name
            if not remove_name.endswith('__to_be_deleted'):
                remove_name += '_%s__to_be_deleted_' % time.time()
                os.rename(name, remove_name)
                #logger.debug("Move completed")
        except OSError as err:
            if err.errno != errno.ENOENT:
                logger.debug("rmrf Err: %s" % err)
                logger.debug("name: %s" % name)
                raise
            return

        for sfn in os.listdir(remove_name):
            try:
                rmrf(os.path.join(remove_name, sfn), count)
            except OSError as err:
                if err.errno == errno.EBUSY:
                    logger.debug("rmrf Remove err: %s" % err)
                    logger.debug("name: %s" % remove_name)
                    ## Sleep 2 sec and try again
                    time.sleep(2.)
                    rmrf(os.path.join(remove_name, sfn), count+1)
        try:
            os.removedirs(remove_name)
        except OSError as err:
            if err.errno == errno.ENOTEMPTY:
                rmrf(remove_name, count+1)
            elif err.errno != errno.ENOENT:
                logger.debug("%s" % err)
                raise
            return
    else:
        try:
            remove_name = name + "_" + str(time.time()) + '__to_be_deleted_'
            os.rename(name, remove_name)
        except OSError as err:
            if err.errno not in [errno.ENOENT, errno.EBUSY]:
                raise
            logger.debug("rmrf Move err: %s" % err)
            logger.debug("name: %s" % name)
            if err.errno == errno.EBUSY:
                rmrf(name, count+1)
            return

        try:
            os.remove(remove_name)
        except OSError as err:
            if err.errno != errno.ENOENT:
                logger.debug("%s" % err)
                logger.debug("name: %s" % remove_name)
                raise
            return


class GangaRepositoryLocal(GangaRepository):

    """GangaRepository Local"""

    def __init__(self, registry):
        """
        Initialize a Repository from within a Registry and keep a reference to the Registry which 'owns' it
        Args:
            Registry (Registry): This is the registry which manages this Repo
        """
        super(GangaRepositoryLocal, self).__init__(registry)
        self.dataFileName = "data"
        self.sub_split = "subjobs"
        self.root = os.path.join(self.registry.location, "6.0", self.registry.name)
        self.lockroot = os.path.join(self.registry.location, "6.0")
        self.saved_paths = {}
        self.saved_idxpaths = {}
        self._cache_load_timestamp = {}
        self.printed_explanation = False
        self._fully_loaded = {}

    def startup(self):
        """ Starts a repository and reads in a directory structure.
        Raise RepositoryError"""
        self._load_timestamp = {}

        # New Master index to speed up loading of many, MANY files
        self._cache_load_timestamp = {}
        self._cached_cat = {}
        self._cached_cls = {}
        self._cached_obj = {}
        self._master_index_timestamp = 0

        self.known_bad_ids = []
        if "XML" in self.registry.type:
            self.to_file = xml_to_file
            self.from_file = xml_from_file
        elif "Pickle" in self.registry.type:
            self.to_file = pickle_to_file
            self.from_file = pickle_from_file
        else:
            raise RepositoryError(self, "Unknown Repository type: %s" % self.registry.type)
        if getConfig('Configuration')['lockingStrategy'] == "UNIX":
            # First test the UNIX locks are working as expected
            try:
                dry_run_unix_locks(self.lockroot)
            except Exception as err:
                # Locking has not worked, lets raise an error
                logger.error("Error: %s" % err)
                msg="\n\nUnable to launch due to underlying filesystem not working with unix locks."
                msg+="Please try launching again with [Configuration]lockingStrategy=FIXED to start Ganga without multiple session support."
                raise RepositoryError(self, msg)

            # Locks passed test so lets continue
            self.sessionlock = SessionLockManager(self, self.lockroot, self.registry.name)
        elif getConfig('Configuration')['lockingStrategy'] == "FIXED":
            self.sessionlock = FixedLockManager(self, self.lockroot, self.registry.name)
        else:
            raise RepositoryError(self, "Unable to launch due to unknown file-locking Strategy: \"%s\"" % getConfig('Configuration')['lockingStrategy'])
        self.sessionlock.startup()
        # Load the list of files, this time be verbose and print out a summary
        # of errors
        self.update_index(True, True)
        logger.debug("GangaRepositoryLocal Finished Startup")

    def shutdown(self):
        """Shutdown the repository. Flushing is done by the Registry
        Raise RepositoryError
        Write an index file for all new objects in memory and master index file of indexes"""
        from GangaCore.Utility.logging import getLogger
        logger = getLogger()
        logger.debug("Shutting Down GangaRepositoryLocal: %s" % self.registry.name)
        for k in self._fully_loaded:
            try:
                self.index_write(k, True)
            except Exception as err:
                logger.error("Warning: problem writing index object with id %s" % k)
        try:
            self._write_master_cache(True)
        except Exception as err:
            logger.warning("Warning: Failed to write master index due to: %s" % err)
        self.sessionlock.shutdown()

    def get_fn(self, this_id):
        """ Returns the file name where the data for this object id is saved
        Args:
            this_id (int): This is the object id we want the XML filename for
        """
        if this_id not in self.saved_paths:
            self.saved_paths[this_id] = os.path.join(self.root, "%ixxx" % int(this_id * 0.001), "%i" % this_id, self.dataFileName)
        return self.saved_paths[this_id]

    def get_idxfn(self, this_id):
        """ Returns the file name where the data for this object id is saved
        Args:
            this_id (int): This is the object id we want the index filename for
        """
        if this_id not in self.saved_idxpaths:
            self.saved_idxpaths[this_id] = os.path.join(self.root, "%ixxx" % int(this_id * 0.001), "%i.index" % this_id)
        return self.saved_idxpaths[this_id]

    def index_load(self, this_id):
        """ load the index file for this object if necessary
            Loads if never loaded or timestamp changed. Creates object if necessary
            Returns True if this object has been changed, False if not
            Raise IOError on access or unpickling error 
            Raise OSError on stat error
            Raise PluginManagerError if the class name is not found
        Args:
            this_id (int): This is the id for which we want to load the index file from disk
        """
        #logger.debug("Loading index %s" % this_id)
        fn = self.get_idxfn(this_id)
        # index timestamp changed
        fn_ctime = os.stat(fn).st_ctime
        cache_time = self._cache_load_timestamp.get(this_id, 0)
        if cache_time != fn_ctime:
            logger.debug("%s != %s" % (cache_time, fn_ctime))
            try:
                with open(fn, 'rb') as fobj:
                    cat, cls, cache = pickle_from_file(fobj)[0]
            except EOFError:
                pass
            except Exception as x:
                logger.warning("index_load Exception: %s" % x)
                raise IOError("Error on unpickling: %s %s" %(getName(x), x))
            if this_id in self.objects:
                obj = self.objects[this_id]
                setattr(obj, "_registry_refresh", True)
            else:
                try:
                    obj = self._make_empty_object_(this_id, cat, cls)
                except Exception as err:
                    raise IOError('Failed to Parse information in Index file: %s. Err: %s' % (fn, err))
            this_cache = obj._index_cache
            this_data = this_cache if this_cache else {}
            for k, v in cache.items():
                this_data[k] = v
            #obj.setNodeData(this_data)
            obj._index_cache = cache
            self._cache_load_timestamp[this_id] = fn_ctime
            self._cached_cat[this_id] = cat
            self._cached_cls[this_id] = cls
            self._cached_obj[this_id] = cache
            return True
        elif this_id not in self.objects:
            self.objects[this_id] = self._make_empty_object_(this_id, self._cached_cat[this_id], self._cached_cls[this_id])
            self.objects[this_id]._index_cache = self._cached_obj[this_id]
            setattr(self.objects[this_id], '_registry_refresh', True)
            return True
        else:
            logger.debug("Doubly loading of object with ID: %s" % this_id)
            logger.debug("Just silently continuing")
        return False

    def index_write(self, this_id, shutdown=False):
        """ write an index file for this object (must be locked).
            Should not raise any Errors,
        Args:
            this_id (int): This is the index for which we want to write the index to disk
            shutdown (bool): True causes this to always be written regardless of any checks"""
        if this_id in self.incomplete_objects:
            return
        logger.debug("Writing index: %s" % this_id)
        obj = self.objects[this_id]
        try:
            ifn = self.get_idxfn(this_id)
            new_idx_cache = self.registry.getIndexCache(stripProxy(obj))
            if not os.path.exists(ifn) or shutdown:
                new_cache = new_idx_cache
                with open(ifn, "wb") as this_file:
                    new_index = (obj._category, getName(obj), new_cache)
                    logger.debug("Writing: %s" % str(new_index))
                    pickle_to_file(new_index, this_file)
                self._cached_obj[this_id] = new_cache
                obj._index_cache = {}
            self._cached_obj[this_id] = new_idx_cache
        except IOError as err:
            logger.error("Index saving to '%s' failed: %s %s" % (ifn, getName(err), err))

    def get_index_listing(self):
        """Get dictionary of possible objects in the Repository: True means index is present,
            False if not present
        Raise RepositoryError"""
        try:
            if not os.path.exists(self.root):
                os.makedirs(self.root)
            obj_chunks = [d for d in os.listdir(self.root) if d.endswith("xxx") and d[:-3].isdigit()]
        except OSError as err:
            logger.debug("get_index_listing Exception: %s" % err)
            raise RepositoryError(self, "Could not list repository '%s'!" % (self.root))
        objs = {}  # True means index is present, False means index not present
        for c in obj_chunks:
            try:
                listing = os.listdir(os.path.join(self.root, c))
            except OSError as err:
                logger.debug("get_index_listing Exception: %s")
                raise RepositoryError(self, "Could not list repository '%s'!" % (os.path.join(self.root, c)))
            objs.update(dict([(int(l), False) for l in listing if l.isdigit()]))
            for l in listing:
                if l.endswith(".index") and l[:-6].isdigit():
                    this_id = int(l[:-6])
                    if this_id in objs:
                        objs[this_id] = True
                    else:
                        try:
                            rmrf(self.get_idxfn(this_id))
                            logger.warning("Deleted index file without data file: %s" % self.get_idxfn(this_id))
                        except OSError as err:
                            logger.debug("get_index_listing delete Exception: %s" % err)
        return objs

    def _read_master_cache(self):
        """
        read in the master cache to reduce significant I/O over many indexes separately on startup
        """
        try:
            _master_idx = os.path.join(self.root, 'master.idx')
            if os.path.isfile(_master_idx):
                logger.debug("Reading Master index")
                self._master_index_timestamp = os.stat(_master_idx).st_ctime
                with open(_master_idx, 'rb') as input_f:
                    this_master_cache = pickle_from_file(input_f)[0]
                for this_cache in this_master_cache:
                    if this_cache[1] >= 0:
                        this_id = this_cache[0]
                        self._cache_load_timestamp[this_id] = this_cache[1]
                        self._cached_cat[this_id] = this_cache[2]
                        self._cached_cls[this_id] = this_cache[3]
                        self._cached_obj[this_id] = this_cache[4]
            else:
                logger.debug("Not Reading Master Index")
        except Exception as err:
            GangaCore.Utility.logging.log_unknown_exception()
            logger.debug("Master Index corrupt, ignoring it")
            logger.debug("Exception: %s" % err)
            self._clear_stored_cache()
        finally:
            rmrf(os.path.join(self.root, 'master.idx'))

    def _clear_stored_cache(self):
        """
        clear the master cache(s) which have been stored in memory
        """
        for k, v in self._cache_load_timestamp.items():
            self._cache_load_timestamp.pop(k)
        for k, v in self._cached_cat.items():
            self._cached_cat.pop(k)
        for k, v in self._cached_cls.items():
            self._cached_cls.pop(k)
        for k, v in self._cached_obj.items():
            self._cached_obj.pop(k)

    def _write_master_cache(self, shutdown=False):
        """
        write a master index cache once per 300sec
        Args:
            shutdown (boool): True causes this to be written now
        """
        try:
            _master_idx = os.path.join(self.root, 'master.idx')
            this_master_cache = []
            if os.path.isfile(_master_idx) and not shutdown:
                if abs(self._master_index_timestamp - os.stat(_master_idx).st_ctime) < 300:
                    return

            items_to_save = iter(self.objects.items())
            for k, v in items_to_save:
                if k in self.incomplete_objects:
                    continue
                try:
                    if k in self._fully_loaded:
                        # Check and write index first
                        obj = v#self.objects[k]
                        new_index = None
                        if obj is not None:
                            new_index = self.registry.getIndexCache(stripProxy(obj))

                        if new_index is not None:
                            #logger.debug("k: %s" % k)
                            arr_k = [k]
                            if len(self.lock(arr_k)) != 0:
                                self.index_write(k)
                                self.unlock(arr_k)
                                self._cached_obj[k] = new_index

                except Exception as err:
                    logger.debug("Failed to update index: %s on startup/shutdown" % k)
                    logger.debug("Reason: %s" % err)

            iterables = iter(self._cache_load_timestamp.items())
            for k, v in iterables:
                if k in self.incomplete_objects:
                    continue
                cached_list = []
                cached_list.append(k)
                try:
                    fn = self.get_idxfn(k)
                    if os.path.isfile(fn):
                        time = os.stat(fn).st_ctime
                    else:
                        time = -1
                except OSError as err:
                    logger.debug("_write_master_cache: %s" % err)
                    logger.debug("_cache_load_timestamp: %s" % self._cache_load_timestamp)
                    import errno
                    if err.errno == errno.ENOENT:  # If file is not found
                        time = -1
                    else:
                        raise

                if time > 0:
                    cached_list.append(time)
                    cached_list.append(self._cached_cat[k])
                    cached_list.append(self._cached_cls[k])
                    cached_list.append(self._cached_obj[k])
                    this_master_cache.append(cached_list)

            try:
                with open(_master_idx, 'wb') as of:
                    pickle_to_file(this_master_cache, of)
            except IOError as err:
                logger.debug("write_master: %s" % err)
                try:
                    os.remove(os.path.join(self.root, 'master.idx'))
                except OSError as x:
                    GangaCore.Utility.logging.log_user_exception(True)
        except Exception as err:
            logger.debug("write_error2: %s" % err)
            GangaCore.Utility.logging.log_unknown_exception()

        return

    def updateLocksNow(self):
        """
        Trigger the session locks to all be updated now
        This is useful when the SessionLock is updating either too slowly or has gone to sleep when there are multiple sessions
        """
        self.sessionlock.updateNow()

    def update_index(self, this_id=None, verbose=False, firstRun=False):
        """ Update the list of available objects
        Raise RepositoryError
        TODO avoid updating objects which haven't changed as this causes un-needed I/O
        Args:
            this_id (int): This is the id we want to explicitly check the index on disk for
            verbose (bool): Should we be verbose
            firstRun (bool): If this is the call from the Repo startup then load the master index for perfomance boost
        """
        # First locate and load the index files
        logger.debug("updating index...")
        objs = self.get_index_listing()
        changed_ids = []
        deleted_ids = set(self.objects.keys())
        summary = []
        if firstRun:
            self._read_master_cache()
        logger.debug("Iterating over Items")

        locked_ids = self.sessionlock.locked

        for this_id in objs:
            deleted_ids.discard(this_id)
            # Make sure we do not overwrite older jobs if someone deleted the
            # count file
            if this_id > self.sessionlock.count:
                self.sessionlock.count = this_id + 1
            # Locked IDs can be ignored
            if this_id in locked_ids:
                continue
            # Skip corrupt IDs
            if this_id in self.incomplete_objects:
                continue
            # Now we treat unlocked IDs
            try:
                # if this succeeds, all is well and we are done
                if self.index_load(this_id):
                    changed_ids.append(this_id)
                continue
            except IOError as err:
                logger.debug("IOError: Failed to load index %i: %s" % (this_id, err))
            except OSError as err:
                logger.debug("OSError: Failed to load index %i: %s" % (this_id, err))
            except PluginManagerError as err:
                # Probably should be DEBUG
                logger.debug("PluginManagerError: Failed to load index %i: %s" % (this_id, err))
                # This is a FATAL error - do not try to load the main file, it
                # will fail as well
                summary.append((this_id, err))
                continue

            # this is bad - no or corrupted index but object not loaded yet!
            # Try to load it!
            if not this_id in self.objects:
                try:
                    logger.debug("Loading disk based Object: %s from %s as indexes were missing" % (this_id, self.registry.name))
                    self.load([this_id])
                    changed_ids.append(this_id)
                    # Write out a new index if the file can be locked
                    if len(self.lock([this_id])) != 0:
                        if this_id not in self.incomplete_objects:
                            # If object is loaded mark it dirty so next flush will regenerate XML,
                            # otherwise just go about fixing it
                            if not self.isObjectLoaded(self.objects[this_id]):
                                self.index_write(this_id)
                            else:
                                self.objects[this_id]._setDirty()
                        #self.unlock([this_id])
                except KeyError as err:
                    logger.debug("update Error: %s" % err)
                    # deleted job
                    if this_id in self.objects:
                        self._internal_del__(this_id)
                        changed_ids.append(this_id)
                except (InaccessibleObjectError, ) as x:
                    logger.debug("update_index: Failed to load id %i: %s" % (this_id, x))
                    summary.append((this_id, x))

        logger.debug("Iterated over Items")

        # Check deleted files:
        for this_id in deleted_ids:
            self._internal_del__(this_id)
            changed_ids.append(this_id)
        if len(deleted_ids) > 0:
            logger.warning("Registry '%s': Job %s externally deleted." % (self.registry.name, ",".join(map(str, list(deleted_ids)))))

        if len(summary) > 0:
            cnt = {}
            examples = {}
            for this_id, x in summary:
                if this_id in self.known_bad_ids:
                    continue
                cnt[getName(x)] = cnt.get(getName(x), []) + [str(this_id)]
                examples[getName(x)] = str(x)
                self.known_bad_ids.append(this_id)
                # add object to incomplete_objects
                if not this_id in self.incomplete_objects:
                    logger.error("Adding: %s to Incomplete Objects to avoid loading it again in future" % this_id)
                    self.incomplete_objects.append(this_id)

            for exc, ids in cnt.items():
                logger.error("Registry '%s': Failed to load %i jobs (IDs: %s) due to '%s' (first error: %s)" % (self.registry.name, len(ids), ",".join(ids), exc, examples[exc]))

            if self.printed_explanation is False:
                logger.error("If you want to delete the incomplete objects, you can type:\n")
                logger.error("'for i in %s.incomplete_ids(): %s(i).remove()'\n (then press 'Enter' twice)" % (self.registry.name, self.registry.name))
                logger.error("WARNING!!! This will result in corrupt jobs being completely deleted!!!")
                self.printed_explanation = True
        logger.debug("updated index done")

        if len(changed_ids) != 0:
            isShutdown = not firstRun
            self._write_master_cache(isShutdown)

        return changed_ids

    def add(self, objs, force_ids=None):
        """ Add the given objects to the repository, forcing the IDs if told to.
        Raise RepositoryError
        Args:
            objs (list): GangaObject-s which we want to add to the Repo
            force_ids (list, None): IDs to assign to object, None for auto-assign
        """

        logger.debug("add")

        if force_ids not in [None, []]:  # assume the ids are already locked by Registry
            if not len(objs) == len(force_ids):
                raise RepositoryError(self, "Internal Error: add with different number of objects and force_ids!")
            ids = force_ids
        else:
            ids = self.sessionlock.make_new_ids(len(objs))

        logger.debug("made ids")

        for i in range(0, len(objs)):
            fn = self.get_fn(ids[i])
            try:
                os.makedirs(os.path.dirname(fn))
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise RepositoryError( self, "OSError on mkdir: %s" % (e))
            self._internal_setitem__(ids[i], objs[i])

            # Set subjobs dirty - they will not be flushed if they are not.
            if self.sub_split and hasattr(objs[i], self.sub_split):
                try:
                    sj_len = len(getattr(objs[i], self.sub_split))
                    if sj_len > 0:
                        for j in range(sj_len):
                            getattr(objs[i], self.sub_split)[j]._dirty = True
                except AttributeError as err:
                    logger.debug("RepoXML add Exception: %s" % err)

        logger.debug("Added")

        return ids

    def _safe_flush_xml(self, this_id):
        """
        Flush XML to disk whilst checking for relavent SubJobXMLList which handles subjobs now
        flush for "this_id" in the self.objects list
        Args:
            this_id (int): This is the id of the object we want to flush to disk
        """

        fn = self.get_fn(this_id)
        obj = self.objects[this_id]
        from GangaCore.Core.GangaRepository.VStreamer import EmptyGangaObject
        if not isType(obj, EmptyGangaObject):
            split_cache = None

            has_children = getattr(obj, self.sub_split, False)

            if has_children:

                logger.debug("has_children")

                if hasattr(getattr(obj, self.sub_split), 'flush'):
                    # I've been read from disk in the new SubJobXMLList format I know how to flush
                    getattr(obj, self.sub_split).flush()
                else:
                    # I have been constructed in this session, I don't know how to flush!
                    if hasattr(getattr(obj, self.sub_split)[0], "_dirty"):
                        split_cache = getattr(obj, self.sub_split)
                        for i in range(len(split_cache)):
                            if not split_cache[i]._dirty:
                                continue
                            sfn = os.path.join(os.path.dirname(fn), str(i), self.dataFileName)
                            if not os.path.exists(os.path.dirname(sfn)):
                                logger.debug("Constructing Folder: %s" % os.path.dirname(sfn))
                                os.makedirs(os.path.dirname(sfn))
                            else:
                                logger.debug("Using Folder: %s" % os.path.dirname(sfn))
                            safe_save(sfn, split_cache[i], self.to_file)
                            split_cache[i]._setFlushed()
                    # Now generate an index file to take advantage of future non-loading goodness
                    tempSubJList = SubJobXMLList(os.path.dirname(fn), self.registry, self.dataFileName, False, obj)
                    ## equivalent to for sj in job.subjobs
                    tempSubJList._setParent(obj)
                    job_dict = {}
                    for sj in getattr(obj, self.sub_split):
                        job_dict[sj.id] = stripProxy(sj)
                    tempSubJList._reset_cachedJobs(job_dict)
                    tempSubJList.flush(ignore_disk=True)
                    del tempSubJList

                safe_save(fn, obj, self.to_file, self.sub_split)
                # clean files not in subjobs anymore... (bug 64041)
                for idn in os.listdir(os.path.dirname(fn)):
                    split_cache = getattr(obj, self.sub_split)
                    if idn.isdigit() and int(idn) >= len(split_cache):
                        rmrf(os.path.join(os.path.dirname(fn), idn))
            else:

                logger.debug("not has_children")

                safe_save(fn, obj, self.to_file, "")
                # clean files leftover from sub_split
                for idn in os.listdir(os.path.dirname(fn)):
                    if idn.isdigit():
                        rmrf(os.path.join(os.path.dirname(fn), idn))
            if this_id not in self.incomplete_objects:
                self.index_write(this_id)
        else:
            raise RepositoryError(self, "Cannot flush an Empty object for ID: %s" % this_id)

        if this_id not in self._fully_loaded:
            self._fully_loaded[this_id] = obj

    def flush(self, ids):
        """
        flush the set of "ids" to disk and write the XML representing said objects in self.objects
        NB: This adds the given objects corresponding to ids to the _fully_loaded dict
        Args:
            ids (list): List of integers, used as keys to objects in the self.objects dict
        """
        logger.debug("Flushing: %s" % ids)

        #import traceback
        #traceback.print_stack()
        for this_id in ids:
            if this_id in self.incomplete_objects:
                logger.debug("Should NEVER re-flush an incomplete object, it's now 'bad' respect this!")
                continue
            try:
                logger.debug("safe_flush: %s" % this_id)
                self._safe_flush_xml(this_id)

                self._cache_load_timestamp[this_id] = time.time()
                self._cached_cls[this_id] = getName(self.objects[this_id])
                self._cached_cat[this_id] = self.objects[this_id]._category
                self._cached_obj[this_id] = self.objects[this_id]._index_cache

                try:
                    self.index_write(this_id)
                except:
                    logger.debug("Index write failed")
                    pass

                if this_id not in self._fully_loaded:
                    self._fully_loaded[this_id] = self.objects[this_id]

                subobj_attr = getattr(self.objects[this_id], self.sub_split, None)
                sub_attr_dirty = getattr(subobj_attr, '_dirty', False)
                if sub_attr_dirty:
                    if hasattr(subobj_attr, 'flush'):
                        subobj_attr.flush()

                self.objects[this_id]._setFlushed()

            except (OSError, IOError, XMLFileError) as x:
                raise RepositoryError(self, "Error of type: %s on flushing id '%s': %s" % (type(x), this_id, x))

    def _check_index_cache(self, obj, this_id):
        """
        Checks the index cache of "this_id" against the index cache generated from the "obj"ect
        If there is a problem then the object is unloaded from memory but will not do anything if everything agrees here
        TODO CHECK IF THIS IS VALID GIVEN WE DYNAMICALLY GENERATE INDEX FOR LOADED OBJECTS
        Args:
            obj (GangaObject): This is the object which we've loaded from disk
            this_id (int): This is the object id which is the objects key in the objects dict
        """
        new_idx_cache = self.registry.getIndexCache(stripProxy(obj))
        if new_idx_cache != obj._index_cache:
            logger.debug("NEW: %s" % new_idx_cache)
            logger.debug("OLD: %s" % obj._index_cache)
            # index is wrong! Try to get read access - then we can fix this
            if len(self.lock([this_id])) != 0:
                if this_id not in self.incomplete_objects:
                    # Mark as dirty if loaded, otherwise load and fix
                    if not self.isObjectLoaded(self.objects[this_id]):
                        self.index_write(this_id)
                    else:
                        self.objects[this_id]._setDirty()
                # self.unlock([this_id])

                old_idx_subset = all((k in new_idx_cache and new_idx_cache[k] == v) for k, v in obj._index_cache.items())
                if not old_idx_subset:
                    # Old index cache isn't subset of new index cache
                    new_idx_subset = all((k in obj._index_cache and obj._index_cache[k] == v) for k, v in new_idx_cache.items())
                else:
                    # Old index cache is subset of new index cache so no need to check
                    new_idx_subset = True

                if not old_idx_subset and not new_idx_subset:
                    if not GANGA_SWAN_INTEGRATION:
                        logger.warning("Incorrect index cache of '%s' object #%s was corrected!" % (self.registry.name, this_id))
                    logger.debug("old cache: %s\t\tnew cache: %s" % (obj._index_cache, new_idx_cache))
                    self.unlock([this_id])
            else:
                pass
                # if we cannot lock this, the inconsistency is
                # most likely the result of another ganga
                # process modifying the repo

    def _parse_xml(self, fn, this_id, load_backup, has_children, tmpobj):
        """
        If we must actually load the object from disk then we end up here.
        This replaces the attrs of "objects[this_id]" with the attrs from tmpobj
        If there are children then a SubJobXMLList is created to manage them.
        The fn of the job is passed to the SubbJobXMLList and there is some knowledge of if we should be loading the backup passed as well
        Args:
            fn (str): This is the path to the data file for this object in the XML
            this_id (int): This is the integer key of the object in the self.objects dict
            load_backup (bool): This reflects whether we are loading the backup 'data~' or normal 'data' XML file
            has_children (bool): This contains the result of the decision as to whether this object actually has children
            tmpobj (GangaObject): This contains the object which has been read in from the fn file
        """

        # If this_id is not in the objects add the object we got from reading the XML
        need_to_copy = True
        if this_id not in self.objects:
            self.objects[this_id] = tmpobj
            need_to_copy = False

        obj = self.objects[this_id]

        # If the object was already in the objects (i.e. cache object, replace the schema content wilst avoiding R/O checks and such
        # The end goal is to keep the object at this_id the same object in memory but to make it closer to tmpobj.
        # TODO investigate changing this to copyFrom
        # The temp object is from disk so all contents have correctly passed through sanitising via setattr at least once by now so this is safe
        if need_to_copy:
            for key, val in tmpobj._data.items():
                obj.setSchemaAttribute(key, val)
            for attr_name, attr_val in obj._schema.allItems():
                if attr_name not in tmpobj._data:
                    obj.setSchemaAttribute(attr_name, obj._schema.getDefaultValue(attr_name))

        if has_children:
            logger.debug("Adding children")
            # NB Keep be a SetSchemaAttribute to bypass the list manipulation which will put this into a list in some cases 
            obj.setSchemaAttribute(self.sub_split, SubJobXMLList(os.path.dirname(fn), self.registry, self.dataFileName, load_backup, obj))
        else:
            if obj._schema.hasAttribute(self.sub_split):
                # Infinite loop if we use setattr btw
                def_val = obj._schema.getDefaultValue(self.sub_split)
                if def_val == []:
                    from GangaCore.GPIDev.Lib.GangaList.GangaList import GangaList
                    def_val = GangaList()
                obj.setSchemaAttribute(self.sub_split, def_val)

        from GangaCore.GPIDev.Base.Objects import do_not_copy
        for node_key, node_val in obj._data.items():
            if isType(node_val, Node):
                if node_key not in do_not_copy:
                    node_val._setParent(obj)

        # Check if index cache; if loaded; was valid:
        if obj._index_cache not in [{}]:
            self._check_index_cache(obj, this_id)

        obj._index_cache = {}

        if this_id not in self._fully_loaded:
            self._fully_loaded[this_id] = obj

    def _load_xml_from_obj(self, fobj, fn, this_id, load_backup):
        """
        This is the method which will load the job from fn using the fobj using the self.from_file method and _parse_xml is called to replace the
        self.objects[this_id] with the correct attributes. We also preseve knowledge of if we're being asked to load a backup or not
        Args:
            fobj (file handler): This is the file handler for the fn
            fn (str): fn This is the name of the file which contains the XML data
            this_id (int): This is the key of the object in the objects dict where the output will be stored
            load_backup (bool): This reflects whether we are loading the backup 'data~' or normal 'data' XML file
        """

        b4=time.time()
        tmpobj, errs = self.from_file(fobj)
        a4=time.time()
        logger.debug("Loading XML file for ID: %s took %s sec" % (this_id, a4-b4))

        if len(errs) > 0:
            logger.error("#%s Error(s) Loading File: %s" % (len(errs), fobj.name))
            for err in errs:
                logger.error("err: %s" % err)
            raise InaccessibleObjectError(self, this_id, errs[0])

        logger.debug("Checking children: %s" % str(this_id))
	#logger.debug("Checking in: %s" % os.path.dirname(fn))
	#logger.debug("found: %s" % os.listdir(os.path.dirname(fn)))

        has_children = SubJobXMLList.checkJobHasChildren(os.path.dirname(fn), self.dataFileName)

        logger.debug("Found children: %s" % str(has_children))

        self._parse_xml(fn, this_id, load_backup, has_children, tmpobj)

        if hasattr(self.objects[this_id], self.sub_split):
            sub_attr = getattr(self.objects[this_id], self.sub_split)
            if sub_attr is not None and hasattr(sub_attr, '_setParent'):
                sub_attr._setParent(self.objects[this_id])

        self._load_timestamp[this_id] = os.fstat(fobj.fileno()).st_ctime

        logger.debug("Finished Loading XML")

    def _open_xml_file(self, fn, this_id, _copy_backup=False):
        """
        This loads the XML for the job "this_id" in self.objects using the file "fn" and knowing whether we want the file or the backup by _copy_backup
        Args:
            fn (str): This is the full XML filename for the given id
            this_id (int): This is the key for the object in the objects dict
            _copy_backup (bool): Should we use the backup file 'data~' (True) or the 'data' file (False)
        """
        fobj = None

        has_loaded_backup = False

        try:
            if not os.path.isfile(fn) and _copy_backup:
                if os.path.isfile(fn + '~'):
                    logger.warning("XML File: %s missing, recovering from backup, recent changes may have been lost!" % fn)
                    has_loaded_backup = True
                    try:
                        from shutil import copyfile
                        copyfile(fn+'~', fn)
                    except:
                        logger.warning("Error Recovering the backup file! loading of Job may Fail!")
            fobj = open(fn, "r")
        except IOError as x:
            if x.errno == errno.ENOENT:
                # remove index so we do not continue working with wrong information
                try:
                    # remove internal representation
                    self._internal_del__(this_id)
                    rmrf(os.path.dirname(fn) + ".index")
                except OSError as err:
                    logger.debug("load unlink Error: %s" % err)
                    pass
                raise KeyError(this_id)
            else:
                raise RepositoryError(self, "IOError: %s" % x)
        finally:
            try:
                if os.path.isdir(os.path.dirname(fn)):
                    ld = os.listdir(os.path.dirname(fn))
                    if len(ld) == 0:
                        os.rmdir(os.path.dirname(fn))
                        logger.warning("No job index or data found, removing empty directory: %s" % os.path.dirname(fn))
            except Exception as err:
                logger.debug("load error %s" % err)
                pass

        return fobj, has_loaded_backup

    def load(self, ids, load_backup=False):
        """
        Load the following "ids" from disk
        If we want to load the backup files for these ids then use _copy_backup
        Correctly loaded objects are dirty, Objects loaded from backups for whatever reason are marked dirty
        Args:
            ids (list): The object keys which we want to iterate over from the objects dict
            load_backup (bool): This reflects whether we are loading the backup 'data~' or normal 'data' XML file
        """
        #print("load: %s " % ids)
        #import traceback
        #traceback.print_stack()
        #print("\n")

        logger.debug("Loading Repo object(s): %s" % ids)

        for this_id in ids:

            if this_id in self.incomplete_objects:
                raise RepositoryError(self, "Trying to re-load a corrupt repository id: %s" % this_id)

            fn = self.get_fn(this_id)
            if load_backup:
                has_loaded_backup = True
                fn = fn + "~"
            else:
                has_loaded_backup = False

            try:
                fobj, has_loaded_backup2 = self._open_xml_file(fn, this_id, True)
                if has_loaded_backup2:
                    has_loaded_backup = has_loaded_backup2
            except Exception as err:
                logger.debug("XML load: Failed to load XML file: %s" % fn)
                logger.debug("Error was:\n%s" % err)
                logger.error("Adding id: %s to Corrupt IDs will not attempt to re-load this session" % this_id)
                self.incomplete_objects.append(this_id)
                raise

            try:
                self._load_xml_from_obj(fobj, fn, this_id, load_backup)
            except RepositoryError as err:
                logger.debug("Repo Exception: %s" % err)
                logger.error("Adding id: %s to Corrupt IDs will not attempt to re-load this session" % this_id)
                self.incomplete_objects.append(this_id)
                raise

            except Exception as err:
                
                should_continue = self._handle_load_exception(err, fn, this_id, load_backup)

                if should_continue is True:
                    has_loaded_backup = True
                    continue
                else:
                    logger.error("Adding id: %s to Corrupt IDs will not attempt to re-load this session" % this_id)
                    self.incomplete_objects.append(this_id)
                    raise

            finally:
                fobj.close()

            subobj_attr = getattr(self.objects[this_id], self.sub_split, None)
            sub_attr_dirty = getattr(subobj_attr, '_dirty', False)

            if has_loaded_backup:
                self.objects[this_id]._setDirty()
            else:
                self.objects[this_id]._setFlushed()

            if sub_attr_dirty:
                getattr(self.objects[this_id], self.sub_split)._setDirty()

        logger.debug("Finished 'load'-ing of: %s" % ids)


    def _handle_load_exception(self, err, fn, this_id, load_backup):
        """
        This method does a lot of the handling of an exception thrown from the load method
        We will return True/False here, True if the error can be correctly caught and False if this is terminal and we couldn't load the object
        Args:
            err (exception): This is the original exception loading the XML data from disk
            fn (str): This is the filename which was used to load the file from disk
            this_id (int): This is the key of the object in the objects dict
            load_backup (bool): This reflects whether we are loading the backup 'data~' or normal 'data' XML file
        """
        if isType(err, XMLFileError):
            logger.error("XML File failed to load for Job id: %s" % this_id)
            logger.error("Actual Error was:\n%s" % err)

        if load_backup:
            logger.debug("Could not load backup object #%s: %s" % (this_id, err))
            raise InaccessibleObjectError(self, this_id, err)

        logger.debug("Could not load object #%s: %s" % (this_id, err))

        # try loading backup
        try:
            self.load([this_id], True)
            logger.warning("Object '%s' #%s loaded from backup file - recent changes may be lost." % (self.registry.name, this_id))
            return True
        except Exception as err2:
            logger.debug("Exception when loading backup: %s" % err2 )

        logger.error("XML File failed to load for Job id: %s" % this_id)
        logger.error("Actual Error was:\n%s" % err)

        # add object to incomplete_objects
        if not this_id in self.incomplete_objects:
            logger.error("Loading: %s into incomplete_objects to avoid loading it again in future" % this_id)
            self.incomplete_objects.append(this_id)
            # remove index so we do not continue working with wrong
            # information
            rmrf(os.path.dirname(fn) + ".index")
            raise InaccessibleObjectError(self, this_id, err)

        return False

    def delete(self, ids):
        """
        This is the method to 'delete' an object from disk, it's written in python and starts with the indexes first
        Args:
            ids (list): The object keys which we want to iterate over from the objects dict
        """
        for this_id in ids:
            # First remove the index, so that it is gone if we later have a
            # KeyError
            fn = self.get_fn(this_id)
            try:
                rmrf(os.path.dirname(fn) + ".index")
            except OSError as err:
                logger.debug("Delete Error: %s" % err)
            self._internal_del__(this_id)
            rmrf(os.path.dirname(fn))
            if this_id in self._fully_loaded:
                del self._fully_loaded[this_id]
            if this_id in self.objects:
                del self.objects[this_id]

    def lock(self, ids):
        """
        Request a session lock for the following ids
        Args:
            ids (list): The object keys which we want to iterate over from the objects dict
        """
        return self.sessionlock.lock_ids(ids)

    def unlock(self, ids):
        """
        Unlock (release file locks of) the following ids
        Args:
            ids (list): The object keys which we want to iterate over from the objects dict
        """
        released_ids = self.sessionlock.release_ids(ids)
        if len(released_ids) < len(ids):
            logger.error("The write locks of some objects could not be released!")

    def get_lock_session(self, this_id):
        """get_lock_session(id)
        Tries to determine the session that holds the lock on id for information purposes, and return an informative string.
        Returns None on failure
        Args:
            this_id (int): Get the id of the session which has a lock on the object with this id
        """
        return self.sessionlock.get_lock_session(this_id)

    def get_other_sessions(self):
        """get_session_list()
        Tries to determine the other sessions that are active and returns an informative string for each of them.
        """
        return self.sessionlock.get_other_sessions()

    def reap_locks(self):
        """reap_locks() --> True/False
        Remotely clear all foreign locks from the session.
        WARNING: This is not nice.
        Returns True on success, False on error."""
        return self.sessionlock.reap_locks()

    def clean(self):
        """clean() --> True/False
        Clear EVERYTHING in this repository, counter, all jobs, etc.
        WARNING: This is not nice."""
        self.shutdown()
        try:
            rmrf(self.root)
        except Exception as err:
           logger.error("Failed to correctly clean repository due to: %s" % err)
        self.startup()

    def isObjectLoaded(self, obj):
        """
        This will return a true false if an object has been fully loaded into memory
        Args:
            obj (GangaObject): The object we want to know if it was loaded into memory
        """
        try:
            _id = next(id_ for id_, o in self._fully_loaded.items() if o is obj)
            return True
        except StopIteration:
            return False




from GangaTest.Framework.tests import GangaGPITestCase
from GangaRobot.Lib.Core.CoreSubmitter import CoreSubmitter
from GangaRobot.Lib.Core.CoreFinisher import CoreFinisher
from GangaRobot.Lib.Core.CoreExtractor import CoreExtractor
from GangaRobot.Lib.Core.CoreReporter import CoreReporter
from GangaRobot.Framework import Utility


class TestCore(GangaGPITestCase):

    """Tests of Core actions and hence indirectly of Base actions."""

    def setUp(self):
        """Create test actions."""
        super(TestCore, self).setUp()
        self.runid = Utility.utcid()
        self.submitter = CoreSubmitter()
        #force submitter patterns
        self.submitter.options = {'CoreSubmitter_Patterns':['GangaRobot/old_test/Lib/Core/test-jobs.txt']}
        self.finisher = CoreFinisher()
        #force finisher timeout to 5 mins
        self.finisher.options = {'BaseFinisher_Timeout':300}
        #test extractor fakes save
        class TestCoreExtractor(CoreExtractor):
            def _saveextract(self, runnode, runid):
                self.runnode = runnode #can be accessed in the test
        self.extractor = TestCoreExtractor()
        #test reporter fakes load and save
        class TestCoreReporter(CoreReporter):
            def _loadextract(self, runid):
                return self.runnode #must be set in the test
            def _savereport(self, report, runid):
                self.report = report #can be accessed in the test 
        self.reporter = TestCoreReporter()

    def tearDown(self):
        """Dereference test actions."""
        self.submitter = None
        self.finisher = None
        self.extractor = None
        self.reporter = None 
        super(TestCore, self).tearDown()

    def test_submitter(self):
        """Test submitter submits 3 jobs and adds them to the jobtree."""
        #execute action
        self.submitter.execute(self.runid)
        #check jobs are added to jobtree
        path = Utility.jobtreepath(self.runid)
        js = jobtree.getjobs(path)
        assert len(js) == 3, 'number of jobs added to jobtree path is not 3'
        for j in js:
            assert j.status != 'new', 'job status is new indicating that it may not have been submitted'

    def test_finisher(self):
        """Test finisher waits for jobs to finish."""
        #submit jobs
        self.submitter.execute(self.runid)
        #execute action
        self.finisher.execute(self.runid)
        #get jobs from jobtree
        path = Utility.jobtreepath(self.runid)
        js = jobtree.getjobs(path)
        #check jobs are finished
        for j in js:
            assert j.status == 'completed', 'job status is not completed indicating that it may not have finished'

    def test_extractor(self):
        """Test extractor creates runnode with correct id and 3 jobs."""
        #submit jobs
        self.submitter.execute(self.runid)
        #execute action
        self.extractor.execute(self.runid)
        #get runnode
        runnode = self.extractor.runnode
        #check runnode contains runid
        assert runnode.getvalue('core.id') == self.runid, 'extracted runnode does not contain correct run id'
        #check runnode contains 3 job nodes
        assert len(runnode.getnodes('job')) == 3, 'extracted runnode does not contain 3 job nodes'

    def test_reporter(self):
        """Test reporter creates report with title and lines and converts to text and html."""
        #submit jobs
        self.submitter.execute(self.runid)
        #extract runnode
        self.extractor.execute(self.runid)
        #set runnode to reporter
        self.reporter.runnode = self.extractor.runnode
        #execute action
        self.reporter.execute(self.runid)
        #get report
        report = self.reporter.report
        #check report is created
        assert report is not None, 'no report created'
        #check report has a title
        assert report.title is not None, 'report does not have a title'
        #check report has some lines
        assert len(report.lines) != 0, 'report does not have any lines'
        #check report converts to plain text
        text = str(report)
        assert text is not None, 'str(report) returns None'
        assert len(text) != 0, 'text version of report is empty'
        #check report converts to html
        html = report.tohtml()
        assert html is not None, 'report.tohtml() returns None'
        assert len(text) != 0, 'html version of report is empty'



# File: Magic.py
# Author: K. Harrison
# Created: 060328

"""Module containing IPython magic functions for Ganga"""

__author__ = "K.Harrison <Harrison@hep.phy.cam.ac.uk>"
__date__ = "28 March 2006"
__version__ = "1.0"

from IPython.core.magic import (Magics, magics_class, line_magic)

from GangaCore.Utility.logging import getLogger
from GangaCore.Utility.Runtime import getScriptPath
from GangaCore.Utility.Runtime import getSearchPath

import sys

logger = getLogger(modulename=1)

@magics_class
class GangaMagics(Magics):
    @line_magic
    def ganga(self, parameter_s=''):
        """
        IPython magic function for executing Python scripts in Ganga namespace

        Usage:
           ganga <script> <arguments>

           <script>    - Python script in Ganga search path
           <arguments> - Arguments to be passed to <script>
        """

        # Obtain list of arguments from input parameter string
        argList = parameter_s.split()

        # Determine path to script, using Ganga search rules
        if argList:
            path = getSearchPath()
            script = getScriptPath(argList[0], path)

            # Use mechanism based on that used in magic_run function of IPython
            # for executing script
            if script:
                save_argv = sys.argv
                sys.argv = [script] + argList[1:]
                prog_ns = self.shell.user_ns
                runner = self.shell.safe_execfile
                runner(script, prog_ns, prog_ns)
                sys.argv = save_argv
            else:
                logger.warning("Script '%s' not found in search path '%s'" %
                               (argList[0], path))
        else:
            logger.info(magic_ganga.__doc__)

        return None



from GangaCore.Core.exceptions import SplitterError
from GangaDirac.Lib.Utilities.DiracUtilities import execute
from GangaDirac.Lib.Backends.DiracUtils import result_ok
from GangaCore.Utility.Config import getConfig
from GangaCore.Utility.logging import getLogger
logger = getLogger()


def igroup(iterable, num, leftovers=False):
    '''
    Generator producing sequential groups of size num from input iterable
    '''
    size = len(iterable)
    for i in range(0, size, num):
        if i + num > size and not leftovers:
            return
        yield iterable[i: i + num]


def GangaDiracSplitter(inputs, filesPerJob, maxFiles, ignoremissing):
    """
    Generator that yields a datasets for dirac split jobs
    """

    split_files = []
    i = inputs.__class__()

    if len(inputs.getLFNs()) != len(inputs.files):
        raise SplitterError("Error trying to split dataset using DIRAC backend with non-DiracFile in the inputdata")

    file_replicas = {}

    from GangaCore.Core.GangaThread.WorkerThreads import getQueues

    for i in inputs:
        #logging.debug( "getting metadata: %s" % str(i.lfn) )
        getQueues().add(i.getReplicas)

    logger.info("Requesting LFN replica info")

    # This finds all replicas for all LFNs...
    # This will probably struggle for LFNs which don't exist
    all_lfns = [i.locations for i in inputs]
    while [] in all_lfns:
        import time
        time.sleep(0.5)
        all_lfns = [i.locations for i in inputs]

    logger.info("Got replicas")

    for i in inputs:
        file_replicas[i.lfn] = i.locations
        #logger.info( "%s" % str( i.accessURL() ) )

    logger.debug("found all replicas")

    super_dict = dict()
    for lfn, repz in file_replicas.items():
        sitez = set([])
        for i in repz:
            # print i
            sitez.add(i)
        super_dict[lfn] = sitez

    allSubSets = []
    allChosenSets = {}

    logger.info("Determining overlap")

    import random
    for i in super_dict.keys():

        # Randomly Select 2 SE as the starting point for spliting jobs
        if len(super_dict[i]) > 2:
            req_sitez = set([])
            chosen = random.sample(super_dict[i], 2)
            for s in chosen:
                req_sitez.add(s)
        # Keep the 2 or less SE as the SE of choice
        else:
            req_sitez = set([])
            for s in super_dict[i]:
                req_sitez.add(s)

        allChosenSets[i] = req_sitez

    logger.debug("Found all SE in use")

    Tier1Sites = set([])

    for i in super_dict.keys():

        req_sitez = allChosenSets[i]
        _this_subset = []

        # Starting with i, populate subset with LFNs which have an
        # overlap of at least 2 SE

        for k in super_dict.keys():
            if req_sitez.issubset(super_dict[k]):
                if len(_this_subset) >= filesPerJob:
                    break
                _this_subset.append(str(k))
                super_dict.pop(k)

        if len(_this_subset) > 0:
            allSubSets.append(_this_subset)

    split_files = allSubSets

    logger.info("Created %s subsets" % str(len(split_files)))

    #logger.info( "Split Files: %s" % str(split_files) )

    for dataset in split_files:
        yield dataset



from GangaCore.Core.GangaThread import GangaThread
from GangaCore.Utility.logging import getLogger
from GangaGUI.gui import app
import requests

logger = getLogger()
gui_server = None


# GangaThread for Flask Server
class GUIServerThread(GangaThread):

    def __init__(self, name: str, host, port):
        GangaThread.__init__(self, name=name)
        self.host = host
        self.port = port

    def run(self):
        try:
            logger.info("Starting GUI Server at {}:{}".format(self.host, self.port))
            app.run(host=self.host, port=self.port)
        except Exception as err:
            logger.error("Failed to start GUI Server: {}".format(err))

    def shutdown(self):
        res = requests.post("http://localhost:{port}/shutdown".format(port=self.port))
        logger.info("{} - {}".format(res.status_code, res.text))


def start_gui(host: str = "localhost", port: int = 5000):
    """Start GUI Flask App on a GangaThread"""
    global gui_server
    gui_server = GUIServerThread("GangaGUI", host, port)
    gui_server.start()


def stop_gui():
    """Stop GUI Flask App on a GangaThread"""
    global gui_server
    if gui_server is not None:
        gui_server.shutdown()


# Use this for starting the server for development purposes
if __name__ == "__main__":
    app.run(debug=True)



"""
Ganga Runtime subsystem: initialization, configuration and creation of runtime objects.
"""


import sys

from .bootstrap import GangaProgram


def setupGanga(argv=sys.argv, interactive=True):
    # Process options given at command line and in configuration file(s)
    # Perform environment setup and bootstrap
    import GangaCore.Runtime
    GangaCore.Runtime._prog = GangaProgram(argv=argv)
    GangaCore.Runtime._prog.parseOptions()
    GangaCore.Runtime._prog.configure()
    GangaCore.Runtime._prog.initEnvironment()
    GangaCore.Runtime._prog.bootstrap(GangaCore.Runtime._prog.interactive)
    GangaCore.Runtime._prog.new_user_wizard(interactive)




################################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: PrimeFactorizerSplitter.py,v 1.1 2008-07-17 16:41:37 moscicki Exp $
################################################################################

from GangaCore.GPIDev.Schema import *
from GangaCore.Utility.Config import getConfig
from GangaCore.GPIDev.Lib.File import File
from GangaCore.GPIDev.Adapters.ISplitter import ISplitter

class PrimeFactorizerSplitter(ISplitter):
    """Job splitter for prime number factorization"""
    
    _name = "PrimeFactorizerSplitter"
    _schema = Schema(Version(1,0), {
        'numsubjobs': SimpleItem(defvalue=1,sequence=0, typelist=['int'], doc="Number of subjobs")
        } )

    ### Splitting based on numsubjobs
    def split(self,job):
        from GangaCore.GPIDev.Lib.Job import Job
        subjobs = []
        primeTables = job.inputdata.get_dataset()

        ## avoid creating jobs with nothing to do
        if self.numsubjobs > len(primeTables):
            self.numsubjobs = len(primeTables)

        ## create subjobs
        for i in range(self.numsubjobs):
            j = Job()
            j.application   = job.application
            j.inputdata     = job.inputdata 
            j.inputdata.table_id_lower = 1 
            j.inputdata.table_id_upper = 1
            j.outputdata    = job.outputdata 
            j.inputsandbox  = job.inputsandbox
            j.outputsandbox = job.outputsandbox
            j.backend       = job.backend
            subjobs.append(j)

        ## chunksize of each subjob
        chunksize = len(primeTables) / self.numsubjobs

        offset = 0
        for i in range(len(subjobs)):
            my_chunksize = chunksize
            if len(primeTables) % self.numsubjobs >= i+1: my_chunksize+=1

            ## set lower bound id (inclusive)
            subjobs[i].inputdata.table_id_lower = int(offset+1)
            ## fill subjob with prime tables 
            #for j in range(my_chunksize):
            #    subjobs[i].application.addPrimeTable(primeTables[offset+j])
            offset += my_chunksize
            ## set upper  bound id (inclusive)
            subjobs[i].inputdata.table_id_upper = int(offset)

        return subjobs




from .Sandbox import SandboxError, createPackedInputSandbox, createInputSandbox, getPackedOutputSandbox
from .WNSandbox import getPackedInputSandbox, createOutputSandbox, createPackedOutputSandbox, OUTPUT_TARBALL_NAME, PYTHON_DIR



################################################################################
# GangaND280 Project.
# Dima Vavilov
# Created 27/01/2016
################################################################################
"""@package ND280Control
Ganga module to execute RunAtmPitSim.exe from the atmPitSim package.
"""

from GangaCore.GPIDev.Adapters.IPrepareApp import IPrepareApp
from GangaCore.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from GangaCore.GPIDev.Schema import *

from GangaCore.Utility.Config import getConfig

from GangaCore.GPIDev.Lib.File import *
from GangaCore.GPIDev.Lib.Registry.PrepRegistry import ShareRef
from GangaCore.GPIDev.Base.Proxy import isType
from GangaCore.Core.exceptions import ApplicationConfigurationError

import os, shutil, re, random
from GangaCore.Utility.files import expandfilename
shared_path = os.path.join(expandfilename(getConfig('Configuration')['gangadir']),'shared',getConfig('Configuration')['user'])

from . import ND280Configs

class runND280Kin(IPrepareApp):
    """
    runND280Kin application running RunAtmPitSim.exe from the atmPitSim package
    for generating .kin files.
        app = runND280Kin()

    The required input for this module are:
        app.cmtsetup = '/home/me/myT2KWork/setup.sh'
        app.confopts = {'nd280ver':'v11r31','fluxfile':'/path/360DegOutputNew10MFlux.kin','nev':3500000}

    NOTE: Ganga will run a bash script so the CMT setup script must be in sh/bash.

    It is also possible to pass command line arguments to the executable.
    For example to use '/home/me/tmp' as temporary directory:
        app.args = ['-t','/home/me/tmp']

    """
    _schema = Schema(Version(1,1), {
        'args' : SimpleItem(defvalue=[],typelist=['str','GangaCore.GPIDev.Lib.File.File.File','int'],sequence=1,strict_sequence=0,doc="List of arguments for the executable. Arguments may be strings, numerics or File objects."),
        'cmtsetup' : SimpleItem(defvalue=[],doc='Setup script(s) in bash to set up cmt and the cmt package of the executable.', typelist=['str'],sequence=1,strict_sequence=0),
        'confopts' : SimpleItem(defvalue={},doc='Options for configuration file', typelist=['str']),
        'env' : SimpleItem(defvalue={},typelist=['str'],doc='Environment'),
        'is_prepared' : SimpleItem(defvalue=None, strict_sequence=0, visitable=1, copyable=1, typelist=['type(None)','bool'],protected=0,comparable=1,doc='Location of shared resources. Presence of this attribute implies the application has been prepared.'),
        } )
    _category = 'applications'
    _name = 'runND280Kin'
    _exportmethods = ['prepare']
    _GUIPrefs = [ { 'attribute' : 'args', 'widget' : 'String_List' },
                  { 'attribute' : 'cmtsetup', 'widget' : 'String' },
                  { 'attribute' : 'confopts', 'widget' : 'String' },
                  { 'attribute' : 'env', 'widget' : 'DictOfString' } ]

    _GUIAdvancedPrefs = [ { 'attribute' : 'args', 'widget' : 'String_List' },
                          { 'attribute' : 'cmtsetup', 'widget' : 'String' },
                          { 'attribute' : 'confopts', 'widget' : 'String' },
                          { 'attribute' : 'env', 'widget' : 'DictOfString' } ]


    def __init__(self):
        super(runND280Kin,self).__init__()


    def configure(self,masterappconfig):
        
        args = convertIntToStringArgs(self.args)

        job = self.getJobObject()

        if self.cmtsetup == []:
          raise ApplicationConfigurationError('No cmt setup script given.')

        infiles = job.inputdata.get_dataset_filenames()
        if len(infiles) != 1:
           raise ApplicationConfigurationError('Wrong Dataset values')
        jn = "%08d" % int(infiles[0])
        outConf = ''
        outConf += "# Automatically generated config file\n\n"
        outConf += "/atmt2k/step/outputFileName corsika_atmpitsim_"+jn+".root\n"
        outConf += "/atmt2k/pga/inputFileName "+self.confopts['fluxfile']+"\n"
        outConf += "/atmt2k/pga/isOyamaFlux false\n"
        outConf += "/atmt2k/pga/inputRandSeed "+str(random.randint(1,9999999))+"\n\n"
        outConf += "/run/verbose 0\n"
        outConf += "/event/verbose 0\n"
        outConf += "/tracking/verbose 0\n"
        outConf += "/run/beamOn "+self.confopts['nev']+"\n"

        mac = "corsika_atmpitsim_"+jn+".mac"
        job.getInputWorkspace().writefile(FileBuffer(mac,outConf),executable=0)

        argsStr = ' '.join(args)
        # Create the bash script and put it in input dir.
        script = '#!/bin/bash\n'
        for f in self.cmtsetup:
            script += 'source '+f+'\n'
        script += 'cd '+job.outputdir+'\n'
        script += 'RunAtmPitSim.exe '+argsStr+' '+os.path.join(job.inputdir,mac)+'\n'

        script += 'mac=$(dirname $(which RunAtmPitSim.exe))/../app/ntuple_2_nuance.C\n'
        script += 'cp $mac .\n'
        script += 'root -l -b <<EOF\n'
        script += '.L ntuple_2_nuance.C+\n'
        script += 'totxt("corsika_atmpitsim_'+jn+'");\n'
        script += '.q\n'
        script += 'EOF\n'

        script += 'mv NoMuons.txt NoMuons_'+jn+'.txt\n'
        script += 'mv Config.conf Config_'+jn+'.conf\n'
        job.getInputWorkspace().writefile(FileBuffer('runND280.sh',script),executable=1)

        self._scriptname = job.inputdir+'runND280.sh'

        # Job name given
        job.name = jn


        return (None,None)



config = getConfig('defaults_runND280') #_Properties
# config.options['exe'].type = type(None)


def convertIntToStringArgs(args):

    result = []
    
    for arg in args:
        if isinstance(arg,int):
            result.append(str(arg))
        else:
            result.append(arg)

    return result



class RTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from GangaCore.GPIDev.Adapters.StandardJobConfig import StandardJobConfig

        c = StandardJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)
        return c
        

class LCGRTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from GangaCore.Lib.LCG import LCGJobConfig

        return LCGJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)

class gLiteRTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from GangaCore.Lib.gLite import gLiteJobConfig

        return gLiteJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)

from GangaCore.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers

allHandlers.add('runND280Kin','LSF', RTHandler)
allHandlers.add('runND280Kin','Local', RTHandler)
allHandlers.add('runND280Kin','PBS', RTHandler)
allHandlers.add('runND280Kin','SGE', RTHandler)
allHandlers.add('runND280Kin','Slurm', RTHandler)
allHandlers.add('runND280Kin','Condor', RTHandler)
allHandlers.add('runND280Kin','LCG', LCGRTHandler)
allHandlers.add('runND280Kin','gLite', gLiteRTHandler)
allHandlers.add('runND280Kin','TestSubmitter', RTHandler)
allHandlers.add('runND280Kin','Interactive', RTHandler)
allHandlers.add('runND280Kin','Batch', RTHandler)
allHandlers.add('runND280Kin','Cronus', RTHandler)
allHandlers.add('runND280Kin','Remote', LCGRTHandler)
allHandlers.add('runND280Kin','CREAM', LCGRTHandler)




"""
A simple configuration interface for Ganga packages.

1) Overview

PackageConfig object corresponds to a configuration section in the INI file.
Typically each plugin handler has its own section. Additionally Ganga provides
a somewhat arbitrary number of other configuration packages.

Configuration of the package is done in several phases:
 - phase one: developer defines the hardcoded values in his package (default level);
              addOption
 - phase two: at startup ganga reads config files and set options (session level);
              setSessionValue() is used
 - phase three: at runtime user may modify options via GPI (user level);
              setUserValue() is used for setting, getEffectiveOption() for getting


2) Defining default options for your package:

#MyPackage.py#

import GangaCore.Utility.Config

config = GangaCore.Utility.Config.getConfig('MyPackage')

config.addOption('opt1','x','this is option 1')
config.addOption('opt2',3.0, 'this is option 2')

The  default  values of  the  options  may  be strings,  numbers,other
built-in  types  or  any  GPI  objects names  defined  in  the  global
config_scope dictionary (in this module).

IMPORTANT: choose the type of the default value carefully: session and
user options  will be automatically  converted to the type  implied by
the default value (you may  also override the type checking default in
setDefaultOption)  .   The  conversion   is  done  as  following  (for
setSessionValue and setUserValue):

 - nothing is done (value is assigned 'as-is'):
    - if the default type is a string and the assigned value is a string
    - if there is no default value

 - eval the string value

 - check if the type matches the default type and raise ConfigError in case of mismatch
    - unless the default type is None
 
Typically strings are assigned via setSessionValue() as they are read
from the config file or command line.

Note for bootstrap procedure: it is OK to first set the session option
and  then  to  set  it's  default  value. That  is  to  say  that  the
configuration may  be read from the  config file prior  to loading the
corresponding module which  uses it. It is NOT OK  to set user options
before the end of the bootstrap procedure.


3) Getting effective values and using callback handlers

To get the effective value of an option you may use:

print(config['opt1'])
print(config.getEffectiveOption('opt1'))
print(config.getEffectiveOptions()) # all options

The  effective value  takes  into account  default,  session and  user
settings  and  may change  at  runtime.  In  GPI  users  only get  the
effective values and may only set values at the user level.

You  may also attach  the callback  handlers at  the session  and user
level.     Pre-process   handlers   may    modify   what    has   been
set. Post-process handlers may be used to trigger extra actions.

def pre(opt,val):
    print('always setting a square of the value')
    return val*2

def post(opt,val):
    print('effectively set',val)
    
config.attachUserHandler(pre,post)

4) How does user see all this in GPI ?

There is a GPI wrapper in GangaCore.GPIDev.Lib.Config which:
     - internally uses setUserValue and getEffectiveOption
     - has a more appealing interface


"""

import os
import re
import traceback
from collections import defaultdict
from functools import reduce

from GangaCore.Core.exceptions import GangaException


class ConfigError(GangaException):

    """ ConfigError indicates that an option does not exist or it cannot be set.
    """

    __slots__=('what',)

    def __init__(self, what=''):
        super(ConfigError, self).__init__()
        self.what = what

    def __str__(self):
        return "ConfigError: %s " % self.what

# WARNING: avoid importing logging at the module level in this file
# for example, do not do here: from GangaCore.Utility.logging import logger
# use getLogger() function defined below:

_logger = None


def getLogger():
    import GangaCore.Utility.logging
    global _logger
    if _logger is not None:
        return _logger

    # for the configuration of the logging package itself (in the initial
    # phases) the logging may be disabled
    try:
        _logger = GangaCore.Utility.logging.getLogger()
        return _logger
    except AttributeError as err:
        print("AttributeError: %s" % err)
        # in such a case we return a mock proxy object which ignore all calls
        # such as logger.info()...

        class X(object):

            def __getattr__(self, name):
                def f(*args, **kwds):
                    pass
                return f
        return X()


# All configuration units
allConfigs = {}

# configuration session buffer
# this dictionary contains the value for the options which are defined in the configuration file and which have
# not yet been defined
unknownConfigFileValues = defaultdict(dict)
unknownGangarcFileValues = defaultdict(dict)
unknownUserConfigValues = defaultdict(dict)

def getConfig(name):
    """
    Get an existing PackageConfig.
    Principle is the same as for getLogger() -- the config instances may
    be easily shared between different parts of the program.

    Args:
        name: the name of the config section

    Returns:
        PackageConfig:

    Raises:
        KeyError: if the config is not found
    """

    try:
        return allConfigs[name]
    except KeyError:
        raise KeyError('Config section "[{0}]" not found'.format(name))


def makeConfig(name, docstring, **kwds):
    """
    Create a config package and attach metadata to it. makeConfig() should be called once for each package.
    """

    if _after_bootstrap:
        raise ConfigError('attempt to create a configuration section [%s] after bootstrap' % name)

    try:
        c = allConfigs[name]
        c.docstring = docstring
        for k in kwds:
            setattr(c, k, kwds[k])
    except KeyError:
        c = allConfigs[name] = PackageConfig(name, docstring, **kwds)

    c._config_made = True
    return c

def makeConfigForTest(name, docstring, **kwds):
    """
    Copy of makeConfig(), but without the bootstrap restrictions as Config Sections created for the Tests
    can be added after bootstrap and should not contain any critical config.
    """

    try:
        c = allConfigs[name]
        c.docstring = docstring
        for k in kwds:
            setattr(c, k, kwds[k])
    except KeyError:
        c = allConfigs[name] = PackageConfig(name, docstring, **kwds)

    c._config_made = True
    return c


class ConfigOption(object):

    """ Configuration Option has a name, default value and a docstring.

    Metainformation:
       * type - if not specified, then the type is inferred from the default value, type may be a type object such as type(1), StringType, type(None) or a list
       of such objects - in this case any type in the list is accepted
       * examples - example how to use the option
       * hidden - True => do not show the option at the level of GPI proxy (default False)
       * cfile - False => do not put the option in the generated config file (default True)
       * filter - None => filter the option value when set (session and user levels)
       * typelist - None => a typelist as in GPI schema 
    The configuration option may also define the session_value and default_value. The value property gives the effective value.
    """

    __slots__ = ('name', 'hidden', 'cfile', 'examples', 'filter', 'typelist', 'hasModified', 'default_value', 'docstring', 'type', 'user_value', 'session_value', 'gangarc_value')

    def __init__(self, name):
        self.name = name
        self.hidden = False
        self.cfile = True
        self.examples = None
        self.filter = None
        self.typelist = None
        self.hasModified = False

    def defineOption(self, default_value, docstring, typelist=None, **meta):

        self.default_value = default_value
        self.docstring = docstring
        self.hidden = False
        self.cfile = True
        self.examples = None
        self.filter = None
        self.typelist = typelist

        for m in meta:
            setattr(self, m, meta[m])

        self.convert_type('session_value')
        self.convert_type('gangarc_value')
        self.convert_type('user_value')

    def setSessionValue(self, session_value):
        if not hasattr(self, 'docstring'):
            raise ConfigError('Can\'t set a session value without a docstring!')

        self.hasModified = True

        # try:
        if self.filter:
            session_value = self.filter(self, session_value)

        if hasattr(self, 'session_value'):
            session_value = self.transform_PATH_option(session_value, self.session_value)

        self.session_value = session_value
        self.convert_type('session_value')

    def setUserValue(self, user_value):
        if not hasattr(self, 'docstring'):
            raise ConfigError('Can\'t set a user value without a docstring!')

        self.hasModified = True
        try:
            if self.filter:
                user_value = self.filter(self, user_value)
        except Exception as x:
            logger = getLogger()
            logger.warning('problem with option filter: %s: %s', self.name, x)

        if hasattr(self, 'user_value'):
            user_value = self.transform_PATH_option(user_value, self.user_value)

        if hasattr(self, 'user_value'):
            old_value = self.user_value

        self.user_value = user_value
        try:
            self.convert_type('user_value')
        except Exception as x:
            # rollback if conversion failed
            try:
                self.user_value = old_value
            except NameError:
                del self.user_value
            raise x

    def setGangarcValue(self, gangarc_value):
        if not hasattr(self, 'docstring'):
            raise ConfigError('Can\'t set a gangarc value without a docstring!')

        self.hasModified = True
        try:
            if self.filter:
                gangarc_value = self.filter(self, gangarc_value)
        except Exception as x:
            logger = getLogger()
            logger.warning('problem with option filter: %s: %s', self.name, x)

        if hasattr(self, 'gangarc_value'):
            gangarc_value = self.transform_PATH_option(gangarc_value, self.gangarc_value)

        if hasattr(self, 'gangarc_value'):
            old_value = self.gangarc_value

        self.gangarc_value = gangarc_value
        try:
            self.convert_type('gangarc_value')
        except Exception as x:
            # rollback if conversion failed
            try:
                self.gangarc_value = old_value
            except NameError:
                del self.gangarc_value
            raise x

    def overrideDefaultValue(self, default_value):
        self.hasModified = True
        if hasattr(self, 'default_value'):
            default_value = self.transform_PATH_option(default_value, self.default_value)
        self.default_value = default_value
        self.convert_type('user_value')
        self.convert_type('gangarc_value')
        self.convert_type('session_value')

    def __getattr__(self, name):

        if name == 'value':
            if self.name.endswith('_PATH'):
                values = []

                for n in ['user', 'gangarc', 'session', 'default']:
                    str_val = n+'_value'
                    if hasattr(self, str_val):
                        values.append(getattr(self, str_val))

                if values:
                    returnable = reduce(self.transform_PATH_option, values)
                    return returnable
            else:
                for n in ['user', 'gangarc', 'session', 'default']:
                    str_val = n+'_value'
                    if hasattr(self, str_val):
                        return getattr(self, str_val)

        elif name == 'level':

            for level, name in [(0, 'user'), (1, 'gangarc'), (2, 'session'), (3, 'default')]:
                if hasattr(self, name + '_value'):
                    return level

        raise AttributeError(name)

    def __setattr__(self, name, value):
        if name in ['value', 'level']:
            raise AttributeError('Cannot set "%s" attribute of the option object' % name)

        super(ConfigOption, self).__setattr__(name, value)
        super(ConfigOption, self).__setattr__('hasModified', True)

    def check_defined(self):
        return hasattr(self, 'default_value')

    def transform_PATH_option(self, new_value, current_value):
        return transform_PATH_option(self.name, new_value, current_value)

    def convert_type(self, x_name):
        """ Convert the type of session_value or user_value (referred to by x_name) according to the types defined by the self.
        If the option has not been defined or the x_name in question is not defined, then this method is no-op.
        If conversion cannot be performed (type mismatch) then raise ConfigError.
        """

        try:
            value = getattr(self, x_name)
        except AttributeError:
            return

        logger = getLogger()

        # calculate the cast type, if it cannot be done then the option has not been yet defined (setDefaultValue)
        # in this case do not modify the value
        if self.typelist is not None:
            # in this case cast_type is a list of string dotnames (like for
            # typelist property in schemas)
            cast_type = self.typelist
        else:
            try:
                # in this case cast_type is a single type object
                cast_type = type(self.default_value)
            except AttributeError:
                return

        new_value = value

        optdesc = 'while setting option [.]%s = %s ' % (self.name, value)

        # eval string values only if the cast_type is not exactly a string
        if isinstance(value, str) and cast_type is not str:
            try:
                new_value = eval(value, config_scope)
                logger.debug('applied eval(%s) -> %s (%s)', value, new_value, optdesc)
            except Exception as x:
                logger.debug('ignored failed eval(%s): %s (%s)', value, x, optdesc)

        # check the type of the value unless the cast_type is not NoneType
        logger.debug('checking value type: %s (%s)', cast_type, optdesc)

        def check_type(x, t):
            return isinstance(x, t) or x is t

        # first we check using the same rules for typelist as for the GPI proxy
        # objects
        try:
            import GangaCore.GPIDev.TypeCheck
            type_matched = GangaCore.GPIDev.TypeCheck._valueTypeAllowed(new_value, cast_type, logger)
        except TypeError:  # cast_type is not a list
            type_matched = check_type(new_value, cast_type)

        from GangaCore.Utility.logic import implies
        if not implies(not cast_type is type(None), type_matched):
            raise ConfigError('type mismatch: expected %s got %s (%s)' % (cast_type, type(new_value), optdesc))

        setattr(self, x_name, new_value)

# indicate if the GPI proxies for the configuration have been created
_after_bootstrap = False

# Scope used by eval when reading-in the configuration.
# Symbols defined in this scope will be correctly evaluated. For example, File class adds itself here.
# This dictionary may also be used by other parts of the system, e.g. XML
# repository.
config_scope = {}


class PackageConfig(object):

    """ Package  Config object  represents a  Configuration  Unit (typically
    related to Ganga Packages). It should not be created directly
    but only via the getConfig method.

    PackageConfig has a name which  corresponds to the [name] section in
    the .gangarc  file.  Once initialized the configuration  may only be
    modified by special setUserValues  methods. This will give a chance
    to Ganga  to take  further actions such  as automatic update  of the
    .gangarc file.

    The PackageConfig interface is designed for Ganga Package Developers.
    User oriented interface is available via the GPI.

    meta keywords:
      - is_open : True => new options may be added by the users (default False)
      - cfile : True => section will be generated in the config file
      - hidden: True => section is not visible in the GPI

    """

    __slots__ = ('name', 'options', 'docstring', 'hidden', 'cfile', '_user_handlers', '_session_handlers', 'is_open', '_config_made', 'hasModified', '__dict__')

    def __init__(self, name, docstring, **meta):
        """ Arguments:
         - name may not contain blanks and should be a valid python identifier otherwise ValueError is raised
         meta args have the same meaning as for the ConfigOption:
          - hidden
          - cfile
        """
        self.name = name
        self.options = {}  # {'name':Option('name',default_value,docstring)}
        self.docstring = docstring
        self.hidden = False
        self.cfile = True

        # processing handlers
        self._user_handlers = []
        self._session_handlers = []
        self._gangarc_handlers = []

        self.is_open = False

        for m in meta:
            setattr(self, m, meta[m])

        # sanity check to force using makeConfig()
        self._config_made = False

        self.hasModified = False

    def _addOpenOption(self, name, value):
        self.addOption(name, value, "", override=True)

    def __iter__(self):
        """  Iterate over the effective options. """
        # return self.getEffectiveOptions().__iter__()
        return self.options.__iter__()

    def __getitem__(self, o):
        """ Get the effective value of option o. """
        return self.getEffectiveOption(o)

    def addOption(self, name, default_value, docstring, override=False, typelist=None, **meta):
        """
        Add a new option to the configuration.
        """
        if _after_bootstrap and not self.is_open:
            raise ConfigError('attempt to add a new option [%s]%s after bootstrap' % (self.name, name))

        # has the option already been made
        try:
            option = self.options[name]
        except KeyError:
            option = ConfigOption(name)

        if option.check_defined() and not override:
            logger = getLogger()
            logger.warning('attempt to add again the option [%s]%s (ignored)', self.name, name)
            return

        option.defineOption(default_value, docstring, typelist, **meta)
        self.options[option.name] = option

        # is it in the list of unknown options from the standard config files
        try:
            conf_value = unknownConfigFileValues[self.name]
        except KeyError:
            msg = "Error getting ConfigFileValue Option: %s" % self.name
            if locals().get('logger') is not None:
                locals().get('logger').debug("dbg: %s" % msg)
            conf_value = dict()

        if option.name in conf_value:
            session_value = conf_value[option.name]
            try:
                option.setSessionValue(session_value)
                del conf_value[option.name]
            except Exception as err:
                msg = "Error Setting Session Value: %s" % err
                if locals().get('logger') is not None:
                    locals().get('logger').debug("dbg: %s" % msg)
        # is it in the list of unknown options specified by the user at the command line or gangarc
        try:
            conf_value = unknownUserConfigValues[self.name]
        except KeyError:
            msg = "Error getting User Option: %s" % self.name
            if locals().get('logger') is not None:
                locals().get('logger').debug("dbg: %s" % msg)
            conf_value = dict()

        if option.name in conf_value:
            user_value = conf_value[option.name]
            try:
                option.setUserValue(user_value) 
                del conf_value[option.name]
            except Exception as err:
                msg = "Error Setting User Value: %s" % err
                if locals().get('logger') is not None:
                    locals().get('logger').debug("dbg: %s" % msg)

        try:
            conf_value = unknownGangarcFileValues[self.name]
        except KeyError:
            msg = "Error getting gangarc Option: %s" % self.name
            if locals().get('logger') is not None:
                locals().get('logger').debug("dbg: %s" % msg)
            conf_value = dict()

        if option.name in conf_value:
            gangarc_value = conf_value[option.name]
            try:
                option.setGangarcValue(gangarc_value) 
                del conf_value[option.name]
            except Exception as err:
                msg = "Error Setting Gangarc Value: %s" % err
                if locals().get('logger') is not None:
                    locals().get('logger').debug("dbg: %s" % msg)


    def setSessionValue(self, name, value):
        """  Add or  override options  as a  part of  second  phase of
        initialization of  this configuration module (PHASE  2) If the
        default type of the option  is not string, then the expression
        will be evaluated."""

        self.hasModified = True

        logger = getLogger()

        logger.debug('trying to set session option [%s]%s = %s', self.name, name, value)

        for h in self._session_handlers:
            value = h[0](name, value)

        try:
            this_opt = self.options[name]
        except KeyError:
            self.options[name] = ConfigOption(name)
            this_opt = self.options[name]

        this_opt.setSessionValue(value)

        logger.debug('sucessfully set session option [%s]%s = %s', self.name, name, value)

        for h in self._session_handlers:
            try:
                h[1](name, value)
            except Exception as err:
                traceback.print_stack()
                logger.error("h[1]: %s" % h[1])
                logger.error("Error in Setting Session Value!")
                logger.error("Name: %s Value: '%s'" % (name, value))
                logger.error("Err:\n%s" % err)
                raise err

    def setUserValue(self, name, value):
        """ Modify option  at runtime. This  method corresponds to  the user
        action so  the value of  the option is considered  'modified' If
        the  default  type  of  the  option  is  not  string,  then  the
        expression will be evaluated. """

        self.hasModified = True

        logger = getLogger()

        logger.debug('trying to set user option [%s]%s = %s', self.name, name, value)

        for handler in self._user_handlers:
            value = handler[0](name, value)

        self.options[name].setUserValue(value)

        logger.debug('successfully set user option [%s]%s = %s', self.name, name, value)

        for handler in self._user_handlers:
            handler[1](name, value)

    def setGangarcValue(self, name, value):
        """ Modify option  at runtime. This  method corresponds to  the gangarc
        action so  the value of  the option is considered  'modified' If
        the  default  type  of  the  option  is  not  string,  then  the
        expression will be evaluated. """

        self.hasModified = True

        logger = getLogger()

        logger.debug('trying to set user option [%s]%s = %s', self.name, name, value)

        for handler in self._user_handlers:
            value = handler[0](name, value)

        self.options[name].setGangarcValue(value)

        logger.debug('successfully set user option [%s]%s = %s', self.name, name, value)

        for handler in self._gangarc_handlers:
            handler[1](name, value)

    def overrideDefaultValue(self, name, val):
        self.hasModified = True
        self.options[name].overrideDefaultValue(val)

    def revertToSession(self, name):
        self.hasModified = True
        try:
            if hasattr(self.options[name], 'user_value'):
                del self.options[name].user_value
        except KeyError:
            pass

    def revertToDefault(self, name):
        self.hasModified = True
        self.revertToSession(name)
        try:
            if hasattr(self.options[name], 'session_value'):
                del self.options[name].session_value
        except KeyError:
            pass

    def revertToSessionOptions(self):
        self.hasModified = True
        for name in self.options:
            self.revertToSession(name)

    def revertToDefaultOptions(self):
        self.hasModified = True
        self.revertToSessionOptions()
        for name in self.options:
            self.revertToDefault(name)

    def getEffectiveOptions(self):
        eff = {}
        for name in self.options:
            eff[name] = self.options[name].value
        return eff

    def getEffectiveOption(self, name):
        try:
            return self.options[name].value
        except KeyError:
            raise ConfigError('option "%s" does not exist in "%s"' % (name, self.name))

    def getEffectiveLevel(self, name):
        """ Return 0 if option is effectively set at the user level, 1
        if at session level or 2 if at default level """
        try:
            return self.options[name].level
        except KeyError:
            raise ConfigError('option "%s" does not exist in "%s"' % (name, self.name))

    def attachUserHandler(self, pre, post):
        """ Attach a user handler:
        - pre(name,x) will be always called before setUserValue(name,x)
        - post(name,x2) will be always called after setUserValue(name,x)

        pre() acts as a filter for the value of the option: its return value (x2) will be set
        Before setting the value will be evaluated (unless a default value type is a string)
        post() will get x2 as the option value.

        It is OK to give None for pre or post. For example:
           config.attachUserHandler(None,post) attaches only the post handler. """

        if pre is None:
            pre = lambda opt, val: val
        if post is None:
            post = lambda opt, val: None

        self._user_handlers.append((pre, post))

    def attachSessionHandler(self, pre, post):
        """See attachUserHandler(). """
        # FIXME: this will NOT always work and should be redesigned, see
        # ConfigOption.filter
        if pre is None:
            pre = lambda opt, val: val
        if post is None:
            post = lambda opt, val: None

        self._session_handlers.append((pre, post))


    def attachGangarcHandler(self, pre, post):
        """See attachUserHandler(). """
        # FIXME: this will NOT always work and should be redesigned, see
        # ConfigOption.filter
        if pre is None:
            pre = lambda opt, val: val
        if post is None:
            post = lambda opt, val: None

        self._gangarc_handlers.append((pre, post))

    def deleteUndefinedOptions(self):
        for o in self.options.keys():
            if not self.options[o].check_defined():
                del self.options[o]

try:
    import configparser
    GangaConfigParser = configparser.ConfigParser
except ImportError:
    # For Python 3
    import configparser as ConfigParser
    GangaConfigParser = configparser.ConfigParser


def make_config_parser(system_vars):
    cfg = GangaConfigParser()
    cfg.optionxform = str  # case sensitive
    cfg.defaults().update(system_vars)
    return cfg


def transform_PATH_option(name, new_value, current_value):
    """
    Return the new value of the option 'name' taking into account special rules for PATH-like variables:

       A variable is PATH-like if the name ends in _PATH.
       Return 'new_value:current_value' unless new_value starts in ':::' or current_value is None.
       In that case return new_value.

       Example:
       if a name of a option terminates in _PATH then the value will not be overriden but
       appended:
       file1.ini:
         ANY_PATH = x
       file2.ini:
         ANY_PATH = y

       result of the merge is: ANY_PATH = x:y

       If you want to override the path you should use :::path, for example:
       file1.ini:
         ANY_PATH = x
       file2.ini
         ANY_PATH = :::y
       result of the merge is: ANY_PATH = y

    For other variables just return the new_value.
    """

    PATH_ITEM = '_PATH'
    if name[-len(PATH_ITEM):] == PATH_ITEM:
        logger = getLogger()
        logger.debug('PATH-like variable: %s %s %s', name, new_value, current_value)
        if current_value is None:
            ret_value = new_value
        elif new_value[:3] != ':::':
            logger.debug('Prepended %s to PATH-like variable %s', new_value, name)
            ret_value = new_value + ':' + current_value
            new_value = ""
        else:
            logger.debug('Resetting PATH-like variable %s to %s', name, new_value)
            ret_value = new_value  # [3:]
            new_value = ":::"

        for tok in ret_value.strip(":").split(":"):
            if new_value.find(tok) == -1 or tok == "":
                new_value += "%s:" % tok

    return new_value


def read_ini_files(filenames, system_vars):
    """ Return  a ConfigParser object  which contains  all options  from the
    sequence of files (which are parsed from left-to-right).
    Apply special rules for PATH-like variables - see transform_PATH_option() """

    import GangaCore.Utility.logging

    logger = getLogger()

    logger.debug('reading ini files: %s', filenames)

    main = make_config_parser(system_vars)

    # load all config files and apply special rules for PATH-like variables
    # note: main.read(filenames) cannot be used because of that

    if isinstance(filenames, str):
        filenames = [filenames]

    for f in filenames:
        if f is None or f == '':
            continue
        cc = make_config_parser(system_vars)
        logger.info('reading config file %s', f)
        try:
            with open(f) as file_f:
                cc.read_file(file_f)
        except Exception as x:
            logger.warning('Exception reading config file %s', x)

        for sec in cc.sections():
            if not main.has_section(sec):
                main.add_section(sec)

            for name in cc.options(sec):
                try:
                    value = cc.get(sec, name)
                except (configparser.InterpolationMissingOptionError, configparser.InterpolationSyntaxError) as err:
                    logger.debug("Parse Error!:\n  %s" % err)
                    value = cc.get(sec, name, raw=True)
                    #raise err

                for localvar in re.finditer('\$\{[^${}]*\}', value):
                    localvarstripped = re.sub(r'[^\w]', '', localvar.group(0))
                    try:
                        value = value.replace(localvar.group(0), cc.get(sec, localvarstripped))
                    except Exception as err:
                        GangaCore.Utility.logging.log_unknown_exception()
                        logger.debug('The variable \"' + localvarstripped + '\" is referenced but not defined in the ')
                        logger.debug('[' + sec + '] configuration section of ' + f)
                        logger.debug("err: %s" % err)

                # do not put the DEFAULTS into the sections (no need)
                if name in cc.defaults():
                    continue

                # special rules (NOT APPLIED IN DEFAULT SECTION):

                try:
                    current_value = main.get(sec, name)
                except configparser.NoOptionError:
                    current_value = None
                except (configparser.InterpolationMissingOptionError, configparser.InterpolationSyntaxError) as err:
                    logger.debug("Parse Error!:\n  %s" % err)
                    logger.debug("Failed to expand, Importing value %s:%s as raw" % (sec, name))
                    current_value = main.get(sec, name, raw=True)
                    current_value = current_value.replace('%', '%%')
                    #raise err

                value = transform_PATH_option(name, value, current_value)

                from GangaCore.Utility.Config import expandgangasystemvars
                value = expandgangasystemvars(None, value)
                # check for the use of environment vars
                re.search('\$\{[^${}]*\}', value)     # matches on ${...}
                for envvar in re.finditer('\$\$[^${}]*\$\$', value):
                    # yeah, if the same variable appears more than once, we'll look it up in the
                    # environment more than once too...but that's not too
                    # arduous.
                    envvar = envvar.group(0)
                    envvarclean = envvar.strip('$')

                    # is env variable
                    logger.debug('looking for ' + str(envvarclean) + ' in the shell environment')
                    if envvarclean in os.environ:
                        envval = os.environ[envvarclean]
                        logger.debug(str(envvarclean) + ' is set as ' + envval + ' in the shell environment')
                        value = value.replace(envvar, envval)
                    else:
                        logger.debug('The configuration file ' + f + ' references an unset environment variable: ' + str(envvarclean))

                # FIXME: strip trailing whitespaces -- SHOULD BE DONE BEFORE IF
                # AT ALL?
                value = value.rstrip()
                value = value.replace('%', '%%')
                try:
                    main.set(sec, name, value)
                except Exception as err:
                    value = value.replace('%', '%%')
                    logger.debug("Error Setting %s" % err)
                    try:
                        main.set(sec, name, value)
                    except Exception as err2:
                        logger.debug("Error setting #2: %s" % err2)
                        raise err

    return main


def setUserValue(config_name, option_name, value):
    """
    Sets the user value for the given config and option.
    If the given config has not already been set the it is added
    to the dict to be added later. The user values supersede the
    session values so are the command line or gangarc options.
    """
    if config_name in allConfigs:
        c = getConfig(config_name)
        if option_name in c.options:
            c.setUserValue(option_name, value)
            return
        if c.is_open:
            c._addOpenOption(option_name, value)
            c.setUserValue(option_name, value)
            return
    # put value in the buffer, it will be removed from the buffer when option
    # is added
    unknownUserConfigValues[config_name][option_name] = value

def setUserValueForTest(config_name, option_name, value):
    """
    Sets the user value for the given config and option for performing Tests.
    Used by (GangaCore/testlib/GangaUnitTest)
    If config_name exists & it's option_name also exists then modifies the option.
    If config_name exists & it's option_name does not exists but options are allowed to be added to config_name (.is_open=True) then adds the option.
    If config_name exists & it's option_name does not exists but options are not allowed to be added to config_name (.is_open=False) then doesn't add the option.
    If config_name does not exists then creates the config_name (with .is_open=True) and add option_name to it. <--- Created for testing purpose
    """

    if config_name in allConfigs:
        c = getConfig(config_name)
        if option_name in c.options:
            c.setUserValue(option_name, value)
            return
        if c.is_open:
            c._addOpenOption(option_name, value)
            c.setUserValue(option_name, value)
            return

    if config_name not in allConfigs:
        c = makeConfigForTest(config_name, '')
        c.is_open = True
        c._addOpenOption(option_name, value)
        c.setUserValue(option_name, value)


def setSessionValue(config_name, option_name, value):
    """
    Sets the session value for the given config and option.
    If the given config has not been set already then it is added
    to the dict to be added later. The session value is superseded by
    the user value.
    """
    if config_name in allConfigs:
        c = getConfig(config_name)
        if option_name in c.options:
            c.setSessionValue(option_name, value)
            return
        if c.is_open:
            c._addOpenOption(option_name, value)
            c.setSessionValue(option_name, value)
            return

    # put value in the buffer, it will be removed from the buffer when option
    # is added
    unknownConfigFileValues[config_name][option_name] = value

def setGangarcValue(config_name, option_name, value):
    """
    Sets the gangarc value for the given config and option.
    If the given config has not been set already then it is added
    to the dict to be added later. The session value is superseded by
    the user value.
    """
    if config_name in allConfigs:
        c = getConfig(config_name)
        if option_name in c.options:
            c.setGangarcValue(option_name, value)
            return
        if c.is_open:
            c._addOpenOption(option_name, value)
            c.setGangarcValue(option_name, value)
            return

    # put value in the buffer, it will be removed from the buffer when option
    # is added
    unknownGangarcFileValues[config_name][option_name] = value

def setSessionValuesFromFiles(filenames, system_vars):
    """ Sets session values for all options in all configuration units
    defined in the sequence of config files.  Initialize config parser
    object with system variables (such as GANGA_TOP, GANGA_VERSION and
    alike).  Contrary to standard  config parser behaviour the options
    from the  DEFAULTS section  are not visible  in the  config units.

    At the time of reading the initialization files, the default options in
    the configuration options (default values) may have not yet been defined.
    """
    gangarcFile = []
    for filename in filenames:
        if 'gangarc' in filename:
            gangarcFile.append(filename)
            filenames.remove(filename)

    #First take the gangarc values
    grcCfg = read_ini_files(gangarcFile, system_vars)

    for name in grcCfg.sections():
        for o in grcCfg.options(name):
            # Important: do not put the options from the DEFAULTS section into
            # the configuration units!
            if o in grcCfg.defaults():
                continue
            try:
                v = grcCfg.get(name, o)
            except (configparser.InterpolationMissingOptionError, configparser.InterpolationSyntaxError) as err:
                logger = getLogger()
                logger.debug("Parse Error!:\n  %s" % err)
                logger.warning("Can't expand the config file option %s:%s, treating it as raw" % (name, o))
                v = grcCfg.get(name, o, raw=True)
            setGangarcValue(name, o, v)


    #Now the others

    cfg = read_ini_files(filenames, system_vars)

    for name in cfg.sections():
        for o in cfg.options(name):
            # Important: do not put the options from the DEFAULTS section into
            # the configuration units!
            if o in cfg.defaults():
                continue
            try:
                v = cfg.get(name, o)
            except (configparser.InterpolationMissingOptionError, configparser.InterpolationSyntaxError) as err:
                logger = getLogger()
                logger.debug("Parse Error!:\n  %s" % err)
                logger.warning("Can't expand the config file option %s:%s, treating it as raw" % (name, o))
                v = cfg.get(name, o, raw=True)
            setSessionValue(name, o, v)


def load_user_config(filename, system_vars):
    logger = getLogger()
    if not os.path.exists(filename):
        return
    new_cfg = read_ini_files(filename, system_vars)
    for name in new_cfg.sections():
        try:
            current_cfg_section = getConfig(name)
        except KeyError:
            continue

        if not current_cfg_section.options:  # if this section does not exists
            # supressing these messages as depending on what stage of the bootstrap.py you
            # call the function more or less of the default options have been loaded
            # currently calling after initialise() could call after bootstrap()
            logger.debug("Section '%s' defined in '%s' is not valid exists and will be removed" % (name, filename))
            continue

        for o in new_cfg.options(name):
            if o not in current_cfg_section.options:
                logger.warning("Option '[%s] %s' defined in '%s' is not valid and will be removed" % (name, o, filename))
                continue
            try:
                v = new_cfg.get(name, o)
            except (configparser.InterpolationMissingOptionError, configparser.InterpolationSyntaxError) as err:
                logger.debug("Parse Error!:\n  %s" % err)
                logger.debug("Failed to expand %s:%s, loading it as raw" % (name, o))
                v = new_cfg.get(name, o, raw=True)
            current_cfg_section.setGangarcValue(o, v)


# KH 050725: Add possibility to overwrite at run-time option set in
#            configuration file
#            => useful for executables where particular actions need
#               to be forced or suppressed
def setConfigOption(section="", item="", value=""):
    """Function to overwrite option values set in configuration file:

       Arguments:
          section - Name of relevant section within configuration file
          item    - Item for which value is to be changed
          value   - Value to be assigned

       Function needs to be called after configuration file has been parsed.

       Return value: None"""

    if bool(section) and bool(item):
        try:
            config = getConfig(section)
            if item in config.getEffectiveOptions():
                config.setSessionValue(item, value)
        except Exception as err:
            getLogger().debug("Error setting Option: %s = %s  :: %s" % (item, value, err))

    return None
# KH 050725: End of addition


def expandConfigPath(path, top):
    """ Split the path and return a list, where all relative path components will have top prepended.
        Example: 'A:/B/C::D/E' -> ['top/A','/B/C','top/D/E']
    """
    l = []
    for p in path.split(':'):
        if p:
            if not os.path.isabs(p):
                p = os.path.join(top, p)
            l.append(p)
    return l


def sanityCheck():
    logger = getLogger()
    for c in allConfigs.values():
        if not c._config_made:
            logger.error("sanity check failed: %s: no makeConfig() found in the code", c.name)

    for name in unknownConfigFileValues:
        opts = unknownConfigFileValues[name]
        if name in allConfigs:
            cfg = allConfigs[name]
        else:
            logger.error("unknown configuration section: [%s]", name)
            continue

        if not cfg.is_open:
            if opts:
                logger.error("unknown options [%s]%s", name, ','.join(list(opts.keys())))
        else:
            # add all options for open sections
            for o, v in zip(list(opts.keys()), list(opts.values())):
                cfg._addOpenOption(o, v)
                cfg.setSessionValue(o, v)



#!/usr/bin/env python

import time
import traceback
from threading import Lock
from queue import Empty
from .Algorithm import AlgorithmError
from GangaCore.Core.GangaThread.GangaThread import GangaThread
from GangaCore.Core.GangaThread.MTRunner.Data import DuplicateDataItemError
from GangaCore.Utility.logging import getLogger


class MTRunnerError(Exception):

    """
    Class for general MTRunner errors.
    """

    def __init__(self, message):
        self.message = message


class GangaWorkAgent(GangaThread):

    def __init__(self, runnerObj, name):
        GangaThread.__init__(self, name=name)
        self._runner = runnerObj

    def run(self):

        logger = getLogger()

        while not self.should_stop():

            if self._runner.data.isEmpty():

                if self._runner.keepAlive:
                    # if self.debug:
                    #    logger.debug('data queue is empty, check again in 0.5 sec.')
                    time.sleep(0.5)
                    continue
                else:
                    logger.debug('data queue is empty, stop worker')
                    break
            else:
                try:
                    item = self._runner.data.getNextItem()

                    logger.debug('worker %s get item %s' %
                                 (self.getName(), item))
                    rslt = self._runner.algorithm.process(item)
                    if rslt:
                        self._runner.lock.acquire()
                        self._runner.doneList.append(item)
                        self._runner.lock.release()
                except NotImplementedError:
                    break
                except AlgorithmError:
                    break
                except Empty:
                    logger.debug("Empty Error!")
                    pass
                # General case to print unexpected exceptions
                except Exception:
                    traceback.print_exc()
                    pass

        self.unregister()


class MTRunner(object):

    """
    Class to handle multiple concurrent threads running on the same algorithm. 

    @since: 0.0.1
    @author: Hurng-Chun Lee 
    @contact: hurngchunlee@gmail.com

    The class itself is a thread. To run it; doing the following:

        runner = MTRunner(myAlgorithm, myData)
        runner.start()
        ... you can do something in parallel in your main program ...
        runner.join()

    where 'myAlorithm' and 'myData' are two objects defining your own
    algorithm running on a dataset.
    """

    _attributes = (
        'name', 'algorithm', 'data', 'numThread', 'doneList', 'lock', 'keepAlive')

    def __init__(self, name, algorithm=None, data=None, numThread=10, keepAlive=False):
        """
        initializes the MTRunner object. 

        @since: 0.0.1
        @author: Hurng-Chun Lee 
        @contact: hurngchunlee@gmail.com

        @param algorithm is an Algorithm object defining how to process on the data
        @param data is an Data object defining what to be processed by the algorithm
        """

        if (not algorithm) or (not data):
            raise MTRunnerError('algorithm and data must not be None')

        self.algorithm = algorithm
        self.data = data
        self.numThread = numThread
        self.doneList = []
        self.lock = Lock()
        self.name = name
        self.keepAlive = keepAlive
        self._agents = []
        self.logger = getLogger()

    def getDoneList(self):
        """
        gets the data items that have been processed correctly by the algorithm.
        """
        return self.doneList

    def getResults(self):
        """
        gets the overall results (e.g. output) from the algorithm.
        """
        return self.algorithm.getResults()

    def addDataItem(self, item):
        """
        adds a new data item into the internal queue
        """
        try:
            self.data.addItem(item)
        except DuplicateDataItemError as e:
            self.logger.debug('skip adding new item: %s' % e.message)

    def start(self):
        """
        starts the MTRunner
        """

        for i in range(self.numThread):
            t = GangaWorkAgent(
                runnerObj=self, name='%s_worker_agent_%d' % (self.name, i))
            self._agents.append(t)
            t.start()

    def join(self, timeout=-1):
        """
        joins the worker agents.

        The caller will be blocked until exceeding the timeout or all worker agents finish their jobs.
        """

        # check the number of alive threads
        try:
            t1 = time.time()

            while self.__cnt_alive_threads__() > 0:
                t2 = time.time()

                # break the loop if exceeding the timeout
                if timeout >= 0 and t2 - t1 > timeout:
                    break
                else:
                    # sleep for another 0.5 second
                    time.sleep(0.5)

        except KeyboardInterrupt:
            self.logger.error(
                'Keyboard interruption on MTRunner: %s' % self.name)

    def stop(self, timeout=-1):
        """
        waits worker agents to finish their works.
        """

        # ask all agents to stop
        for agent in self._agents:
            agent.stop()

        self.join(timeout=timeout)

    def __cnt_alive_threads__(self):

        num_alive_threads = 0
        for t in self._agents:
            if t.is_alive():
                num_alive_threads += 1

        return num_alive_threads



from GangaCore.GPIDev.Schema.Schema import Schema, SimpleItem, Version
from GangaCore.GPIDev.Base.Objects import GangaObject
from GangaCore.Utility.logging import getLogger
from GangaCore.Core.GangaRepository.GangaRepository import RepositoryError
from GangaCore.Core.exceptions import GangaException
from GangaCore.GPIDev.Base.Proxy import stripProxy
from GangaCore.Core.GangaRepository.VStreamer import XMLFileError
import errno
import copy
import threading
import shutil
from os import listdir, path, stat

logger = getLogger()

##FIXME There has to be a better way of doing this?
class SJXLIterator(object):
    """Class for iterating over SJXMLList, potentially very unstable, dangerous and only supports looping forwards ever"""

    __slots__ = ('_myCount', '_mySubJobs')

    def __init__(self, theseSubJobs):
        """ Iterator constructor
        Args:
            theseSubJobs (list): sequential list of subjobs we're to iterate over
        """
        self._mySubJobs = theseSubJobs
        self._myCount = 0

    ## NB becomes __next__ in Python 3.x don't know if Python 2.7 has a wrapper here
    def __next__(self):
        if self._myCount < len(self._mySubJobs):
            returnable = self._mySubJobs[self._myCount]
            self._myCount += 1
            return returnable
        else:
            raise StopIteration


class SubJobXMLList(GangaObject):
    """
        SUBJOBXMLLIST class for managing the subjobs so they're loaded only when needed
    """

    _category = 'internal'
    _exportmethods = ['__getitem__', '__len__', '__iter__', 'getAllCachedData', 'values']
    _hidden = True
    _name = 'SubJobXMLList'

    _schema = Schema(Version(1, 0), {})

    def __init__(self, jobDirectory='', registry=None, dataFileName='data', load_backup=False, parent=None):
        """ Constructor for SubjobXMLList
        Args:
            jobDirectory (str): dir on disk which contains subjob folders
            registry (Registry): the registry managing me,
            dataFileName (str): incase it ever changes, normally 'data'
            load_backup (bool): are we using the backpus only/first? This used to be set like this btw
            paret (Job): parent of self after constuction
        """
        super(SubJobXMLList, self).__init__()

        self._jobDirectory = jobDirectory
        self._registry = registry
        self._cachedJobs = {}

        self._dataFileName = dataFileName
        self._load_backup = load_backup

        self._definedParent = None

        self._subjob_master_index_name = "subjobs.idx"

        if jobDirectory == '' and registry is None:
            return

        self._subjobIndexData = {}
        if parent:
            self._setParent(parent)
        self.load_subJobIndex()

        self._cached_filenames = {}
        self._stored_len = []
        # For caching a large list of integers, the key is the length of the list
        self._storedKeys = {}

        # Lock to ensure only one load at a time
        self._load_lock = threading.Lock()


    ## THIS CLASS MAKES USE OF THE INTERNAL CLASS DICTIONARY ONLY!!!
    ## THIS CLASS DOES NOT MAKE USE OF THE SCHEMA TO STORE INFORMATION AS TRANSIENT OR UNCOPYABLE
    ## THIS CLASS CONTAINS A LOT OF OBJECT REFERENCES WHICH SHOULD NOT BE DEEPCOPIED!!!
    def __deepcopy__(self, memo=None):
        obj = super(SubJobXMLList, self).__deepcopy__(memo)

        obj._subjobIndexData = copy.deepcopy(self._subjobIndexData, memo)
        obj._jobDirectory = copy.deepcopy(self._jobDirectory, memo)
        obj._registry = self._registry
        obj._dataFileName = copy.deepcopy(self._dataFileName, memo)
        obj._load_backup = copy.deepcopy(self._load_backup, memo)
        obj._cached_filenames = copy.deepcopy(self._cached_filenames, memo)
        obj._stored_len = copy.deepcopy(self._stored_len, memo)

        ## Manually define unsafe/uncopyable objects
        obj._definedParent = None
        obj._cachedJobs = {}
        return obj

    def _reset_cachedJobs(self, obj):
        """Hard reset function. Not really to be used externally without great care
        Args:
            obj (dict): This is the new dictonary of subjob Job objects with sequential integer keys
        """
        self._cachedJobs = obj

    def isLoaded(self, subjob_id):
        """Has the subjob been loaded? True/False
        Args:
            subjob_id (int): This is the id of the job that we're interested in
        """
        return subjob_id in self._cachedJobs

    def load_subJobIndex(self):
        """Load the index from all sujobs ynto _subjobIndexData or empty it is an error occurs"""
        index_file = path.join(self._jobDirectory, self._subjob_master_index_name )
        if path.isfile( index_file ):
            index_file_obj = None
            try:
                from GangaCore.Core.GangaRepository.PickleStreamer import from_file

                try:
                    index_file_obj = open(index_file, "rb" )
                    self._subjobIndexData = from_file( index_file_obj )[0]
                except IOError as err:
                    self._subjobIndexData = None
                    self._setDirty()

                if self._subjobIndexData is None:
                    self._subjobIndexData = {}
                else:
                    for subjob_id in self._subjobIndexData:
                        index_data = self._subjobIndexData.get(subjob_id)
                        ## CANNOT PERFORM REASONABLE DISK CHECKING ON AFS
                        ## SLOW FILE ACCESS WRITE AND METADATA MEANS FILE DATA DOES NOT MATCH MOD TIME
                        #if index_data is not None and 'modified' in index_data:
                        #    mod_time = index_data['modified']
                        #    disk_location = self.__get_dataFile(str(subjob_id))
                        #    disk_time = stat(disk_location).st_ctime
                        #    diff = disk_time - mod_time
                        #    if disk_time > mod_time and (diff*diff < 9.):
                        #        logger.warning("objs: %s" % self._cachedJobs.keys())
                        #        logger.warning("%s != %s" % (mod_time, disk_time))
                        #        logger.warning("SubJob: %s has been modified, re-loading" % (subjob_id))
                        #        new_data = self._registry.getIndexCache( self.__getitem__(subjob_id) )
                        #        self._subjobIndexData[subjob_id] = new_data
                        #        break
                        #else:
                        if index_data is None:
                            logger.warning("Cannot find subjob index %s, rebuilding" % subjob_id)
                            new_data = self._registry.getIndexCache( self.__getitem__(subjob_id) )
                            self._subjobIndexData[subjob_id] = new_data
                            continue
                        #self._subjobIndexData = {}
            except Exception as err:
                logger.debug( "Subjob Index file open, error: %s" % err )
                self._subjobIndexData = {}
                self._setDirty()
            finally:
                if index_file_obj is not None:
                    index_file_obj.close()
                if self._subjobIndexData is None:
                    self._subjobIndexData = {}
        else:
            self._setDirty()
        return

    def write_subJobIndex(self, ignore_disk=False):
        """interface for writing the index which captures errors and alerts the user vs throwing uncaught exception
        Args:
            ignore_disk (bool): Optional flag to force the class to ignore all on-disk data when flushing
        """
        try:
            self.__really_writeIndex(ignore_disk)
        ## Once It's known what te likely exceptions here are they'll be added
        except (IOError,) as err:
            logger.debug("Can't write Index. Moving on as this is not essential to functioning it's a performance bug")
            logger.debug("Error: %s" % err)

    def __really_writeIndex(self, ignore_disk=False):
        """Do the actual work of writing the index for all subjobs
        Args:
            ignore_disk (bool): Optional flag to force the class to ignore all on-disk data when flushing
        """

        all_caches = {}
        if ignore_disk:
            range_limit = list(self._cachedJobs.keys())
        else:
            range_limit = list(range(len(self)))

        for sj_id in range_limit:
            if sj_id in self._cachedJobs:
                this_cache = self._registry.getIndexCache(self.__getitem__(sj_id))
                all_caches[sj_id] = this_cache
                disk_location = self.__get_dataFile(sj_id)
                all_caches[sj_id]['modified'] = stat(disk_location).st_ctime
            else:
                if sj_id in self._subjobIndexData:
                    all_caches[sj_id] = self._subjobIndexData[sj_id]
                else:
                    this_cache = self._registry.getIndexCache(self.__getitem__(sj_id))
                    all_caches[sj_id] = this_cache
                    disk_location = self.__get_dataFile(sj_id)
                    all_caches[sj_id]['modified'] = stat(disk_location).st_ctime

        try:
            from GangaCore.Core.GangaRepository.PickleStreamer import to_file
            index_file = path.join(self._jobDirectory, self._subjob_master_index_name)
            index_file_obj = open(index_file, "wb")
            to_file(all_caches, index_file_obj)
            index_file_obj.close()
        ## Once I work out what the other exceptions here are I'll add them
        except (IOError,) as err:
            logger.debug("cache write error: %s" % err)

    def __iter__(self):
        """Return iterator for this class"""
        return SJXLIterator(self)

    def __get_dataFile(self, index, force_backup=False):
        """Get the filename for this file (with out without backup '~'. Store already determine combinations in _cached_filenames for speed
        Args:
            index (int): This is the index of the subjob we're interested in
            force_backup (bool): Should we force the loading from the backup XML
        """

        backup_decision = self._load_backup is True or force_backup is True

        index_str = str(index)+"_"+str(backup_decision)

        if index_str in self._cached_filenames:
            return self._cached_filenames[index_str]

        subjob_data = path.join(self._jobDirectory, str(index), self._dataFileName)
        if backup_decision is True:
            subjob_data = subjob_data + '~'

        self._cached_filenames[index_str] = subjob_data
        return subjob_data

    def __len__(self):
        """ return length or lookup the last modified time compare against self._stored_len[0] and if nothings changed return self._stored_len[1]"""
        try:
            this_time = stat(self._jobDirectory).st_ctime
        except OSError:
            return 0

        if len(self._stored_len) == 2:
            last_time = self._stored_len[0]
            if this_time == last_time:
                return self._stored_len[1]

        if not path.isdir( self._jobDirectory ):
            return 0

        subjob_count = SubJobXMLList.countSubJobDirs(self._jobDirectory, self._dataFileName, False)

        if len(self._stored_len) != 2:
            self._stored_len = []
            self._stored_len.append(this_time)
            self._stored_len.append(subjob_count)
        else:
            self._stored_len[0] = this_time
            self._stored_len[1] = subjob_count

        return subjob_count

    def keys(self):
        """Return keys to access subjobs"""
        myLen = len(self)
        if myLen not in self._storedKeys:
            self._storedKeys[myLen] = [i for i in range(len(self))]
        return self._storedKeys[myLen]

    def values(self):
        """Return the actual subjobs"""
        return [self[i] for i in range(0, len(self))]

    def getSafeJob(self):
        """ Return the job object or None, no faffing around with throwing exceptions """
        try:
            job_obj = self.getJobObject()
        except Exception as err:
            logger.debug("Error: %s" % err)
            try:
                job_obj = self._getParent()
            except Exception as err:
                job_obj = None
        return job_obj

    def getMasterID(self):
        """ Return a string corresponding to the parent job ID """
        job_obj = self.getSafeJob()
        if job_obj is not None:
            try:
                fqid = job_obj.getFQID('.')
            except Exception as err:
                try:
                    fqid = job_obj.id
                except:
                    fqid = "unknown"
        else:
            fqid = "unknown"
        return fqid

    def _loadSubJobFromDisk(self, subjob_data):
        """Load the subjob file 'subjob_data' from disk. No Parsing
        Args:
            subjob_data (str): filename for the subjob 'data' file we're interested in
        """
        # For debugging where this was called from to try and push it to as high a level as possible at runtime
        #print("SJXML Load")
        #import traceback
        #traceback.print_stack()
        #print("\n\n\n")
        #import sys
        #sys.exit(-1)
        job_obj = self.getSafeJob()
        if job_obj is not None:
            fqid = self.getMasterID()
            logger.debug( "Loading subjob at: %s for job %s" % (subjob_data, fqid) )
        else:
            logger.debug( "Loading subjob at: %s" % subjob_data )
        sj_file = open(subjob_data, "r")
        return sj_file

    def __call__(self, index):
        """Same as getitem
        Args:
            index (int): The index corresponding to the subjob object we want
        """
        return self.__getitem__(index)

    def __getitem__(self, index):
        """Return a subjob based upon index
        Args:
            index (int): The index corresponding to the subjob object we want
        """
        try:
            return self._getItem(index)
        except (GangaException, IOError, XMLFileError) as err:
            logger.error("CANNOT LOAD SUBJOB INDEX: %s. Reason: %s" % (index, err))
            raise

    def _getItem(self, index):
        """Actual meat of loading the subjob from disk is required, parsing and storing a copy in memory
        (_cached_subjobs) for future use
        Args:
            index (int): The index corresponding to the subjob object we want
        """
        logger.debug("Requesting subjob: #%s" % index)

        if index not in self._cachedJobs:

            logger.debug("Attempting to load subjob: #%s from disk" % index)

            # obtain a lock to make sure multiple loads of the same object don't happen
            with self._load_lock:

                # just make sure we haven't loaded this object already while waiting on the lock
                if index in self._cachedJobs:
                    return self._cachedJobs[index]

                has_loaded_backup = False

                # Now try to load the subjob
                if len(self) < index:
                    raise GangaException("Subjob: %s does NOT exist" % index)
                subjob_data = self.__get_dataFile(str(index))
                try:
                    sj_file = self._loadSubJobFromDisk(subjob_data)
                except (XMLFileError, IOError) as x:
                    logger.warning("Error loading XML file: %s" % x)
                    try:
                        logger.debug("Loading subjob #%s for job #%s from disk, recent changes may be lost" % (index, self.getMasterID()))
                        subjob_data = self.__get_dataFile(str(index), True)
                        sj_file = self._loadSubJobFromDisk(subjob_data)
                        has_loaded_backup = True
                    except (IOError, XMLFileError) as err:
                        logger.debug("Error loading subjob XML:\n%s" % err)

                        if isinstance(x, IOError) and x.errno == errno.ENOENT:
                            raise IOError("Subobject %s not found: %s" % (index, x))
                        else:
                            raise RepositoryError(self,"IOError on loading subobject %s: %s" % (index, x))

                from GangaCore.Core.GangaRepository.VStreamer import from_file

                # load the subobject into a temporary object
                try:
                    loaded_sj = from_file(sj_file)[0]
                except (IOError, XMLFileError) as err:

                    try:
                        logger.warning("Loading subjob #%s for job #%s from backup, recent changes may be lost" % (index, self.getMasterID()))
                        subjob_data = self.__get_dataFile(str(index), True)
                        sj_file = self._loadSubJobFromDisk(subjob_data)
                        loaded_sj = from_file(sj_file)[0]
                        has_loaded_backup = True
                    except (IOError, XMLFileError) as err:
                        logger.debug("Failed to Load XML for job: %s using: %s" % (index, subjob_data))
                        logger.debug("Err:\n%s" % err)
                        raise

                loaded_sj._setParent( self._definedParent )
                if has_loaded_backup:
                    loaded_sj._setDirty()
                else:
                    loaded_sj._setFlushed()
                self._cachedJobs[index] = loaded_sj

        return self._cachedJobs[index]

    def _setParent(self, parentObj):
        """Set the parent of self and any objects in memory we control
        Args:
            parentObj (Job): This is the master job we're hanging these children off in the tree
        """

        if parentObj is not None and hasattr(parentObj, 'getFQID'):
            parent_name = "Job: %s" % parentObj.getFQID('.')
        elif parentObj is not None and hasattr(parentObj, 'id'):
            parent_name = "Job: %s" % parentObj.id
        else:
            parent_name = "None"
        logger.debug('Setting Parent: %s' % parent_name)

        super(SubJobXMLList, self)._setParent(parentObj)

        if self._definedParent is not parentObj:
            self._definedParent = parentObj

        if not hasattr(self, '_cachedJobs'):
            return
        for k in self._cachedJobs:
            if self._cachedJobs[k]._getParent() is not self._definedParent:
                self._cachedJobs[k]._setParent( parentObj )

    def getCachedData(self, index):
        """Get the cached data from the index for one of the subjobs
        Args:
            index (int): index for the subjob we're interested in
        """
        if index > len(self) or index < 0:
            return None

        if index in self._subjobIndexData:
            if self.isLoaded(index):
                return self._registry.getIndexCache( self.__getitem__(index) )
            else:
                return self._subjobIndexData[index]
        else:
            return self._registry.getIndexCache( self.__getitem__(index) )

        return None

    def getAllCachedData(self):
        """Get the cached data from the index for all subjobs"""
        cached_data = []
        #logger.debug("Cache: %s" % self._subjobIndexData)
        if len(self._subjobIndexData) == len(self):
            for i in range(len(self)):
                if self.isLoaded(i):
                    cached_data.append( self._registry.getIndexCache( self.__getitem__(i) ) )
                else:
                    cached_data.append( self._subjobIndexData[i] )
        else:
            for i in range(len(self)):
                cached_data.append(self._registry.getIndexCache( self.__getitem__(i) ) )

        return cached_data

    def getAllSJStatus(self):
        """
        Returns the cached statuses of the subjobs whilst respecting the Lazy loading
        """
        sj_statuses = []
        if len(self._subjobIndexData) == len(self):
            for i in range(len(self)):
                if self.isLoaded(i):
                    sj_statuses.append(self.__getitem__(i).status)
                else:
                    sj_statuses.append(self._subjobIndexData[i]['status'])
        else:
            for i in range(len(self)):
                sj_statuses.append(self.__getitem__(i).status)
        return sj_statuses

    def flush(self, ignore_disk=False):
        """Flush all subjobs to disk using XML methods
        Args:
            ignore_disk (bool): Optional flag to force the class to ignore all on-disk data when flushing
        """
        from GangaCore.Core.GangaRepository.GangaRepositoryXML import safe_save

        from GangaCore.Core.GangaRepository.VStreamer import to_file

        if ignore_disk:
            range_limit = list(self._cachedJobs.keys())
        else:
            range_limit = list(range(len(self)))

        for index in range_limit:
            if index in self._cachedJobs:
                ## If it ain't dirty skip it
                if not self._cachedJobs[index]._dirty:
                    continue

                subjob_data = self.__get_dataFile(str(index))
                subjob_obj = self._cachedJobs[index]

                if subjob_obj is subjob_obj._getRoot():
                    raise GangaException(self, "Subjob parent not set correctly in flush.")

                safe_save( subjob_data, subjob_obj, to_file )

        self.write_subJobIndex(ignore_disk)

    def _setFlushed(self):
        """ Like Node only descend into objects which aren't in the Schema"""
        for index in self._cachedJobs:
            self._cachedJobs[index]._setFlushed()
        super(SubJobXMLList, self)._setFlushed()

    def _private_display(self, reg_slice, this_format, default_width, markup):
        """ This is a private display method which makes use of the display slice as well as knowlede of the wanted format, default_width and markup to be used
        Given it's  display method this returns a displayable string. Given it's tied into the RegistrySlice it's similar to that
        Args:
            reg_slice (RegistrySlice): This is the registry slice which is the context in which this is called
            this_format (str): This is the format used in the registry slice for the formatting of the table
            defult_width (int): default width for a colum as defined in registry slice
            markup (str): This is the markup function used to format the text in the table from registry slice
        """
        ds=""
        for obj_i in self.keys():

            cached_data = self.getCachedData(obj_i)
            colour = reg_slice._getColour(cached_data)

            vals = []
            for item in reg_slice._display_columns:
                display_str = "display:" + str(item)
                #logger.debug("Looking for : %s" % display_str)
                width = reg_slice._display_columns_width.get(item, default_width)
                try:
                    if item == 'fqid':
                        vals.append(str(cached_data[display_str]))
                    else:
                        vals.append(str(cached_data[display_str])[0:width])
                except KeyError as err:
                    logger.debug("_private_display KeyError: %s" % err)
                    vals.append("Unknown")

            ds += markup(this_format % tuple(vals), colour)

        return ds

    @staticmethod
    def checkJobHasChildren(jobDirectory, datafileName):
        """ Return True/False if given (job?) object has children associated with it
        This function will test for the presence of all of the subjob XML in the appropriate folders and will trigger an exception
        if/when some of the xml files are missing. This is subtly different to the default countSubJobDirs behaviour.
        Args:
            jobDirectory (str): name of folder to be examined
            datafileName (str): name of the files containing the xml, i.e. 'data' by convention
        """

        if not path.isdir(jobDirectory):
            return False
        else:
            return bool(SubJobXMLList.countSubJobDirs(jobDirectory, datafileName, True))

    @staticmethod
    def countSubJobDirs(jobDirectory, datafileName, checkDataFiles):
        """ I'm a function which returns a number, my number corresponds to the amount of sequentially listed numerically named folders exiting within 'jobDirectory'
            This (optionally) checks for the existance of all of the XML files. This is useful during a call to 'jobHasChildrenTest' but not when calling __len__ repeatedly
        Args:
            jobDirectory (str): name of folder to be examined
            datafileName (str): name of the files containing the xml, i.e. 'data' by convention
            checkDataFiles (bool): if True check for the existance of all of the data files and check this against the numerically named folders
        """

        jobDirectoryList = listdir(jobDirectory)

        subjob_count=0
        for dir_entry in jobDirectoryList:
            if dir_entry.isdigit():
                sj_dir = path.join(jobDirectory, dir_entry)
                if path.isdir(sj_dir):
                    if checkDataFiles:
                        data_file_path = path.join(sj_dir, datafileName)
                        if path.isfile(data_file_path):
                            subjob_count+=1
                        elif path.isfile(data_file_path+'~'):
                            logger.warning("Reverting to backup due to missing XML: %s" % data_file_path)
                            subjob_count+=1
                    else:
                        subjob_count+=1

        logger.debug("count: %s len: %s" % (subjob_count, len([_folder for _folder in jobDirectoryList if _folder.isdigit()])))

        if subjob_count == len([_folder for _folder in jobDirectoryList if _folder.isdigit()]):
            return subjob_count
        else:
            raise GangaException("Missing subjobs data file in %s" % jobDirectory)




##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: files.py,v 1.1 2008-07-17 16:41:00 moscicki Exp $
##########################################################################

"""
Helper functions for operations on files.
"""

import os
import glob
import stat
import shutil

_stored_expanded_paths = {}
_stored_full_paths = {}

def expandfilename(filename, force=False):
    """expand a path or filename in a standard way so that it may contain ~ and ${VAR} strings"""
    if filename in _stored_expanded_paths:
        expanded_path = _stored_expanded_paths[filename]
    else:
        expanded_path = os.path.expandvars(os.path.expanduser(filename))
        _stored_expanded_paths[filename] = expanded_path
    if os.path.exists(expanded_path) or force:
        return expanded_path

    return filename


def fullpath(path, force=False):
    """expandfilename() and additionally: strip leading and trailing whitespaces and expand symbolic links"""
    if path in _stored_full_paths:
        full_path = _stored_full_paths[path]
    else:
        full_path = os.path.realpath(expandfilename(path.strip(), True))
        _stored_full_paths[path] = full_path
    if os.path.exists(full_path) or force:
        return full_path

    return path


def previous_dir(path, cnt):
    "returns the previous cnt-th directory"
    for i in range(cnt):
        path = os.path.dirname(path)
    return path


def chmod_executable(path):
    "make a file executable by the current user (u+x)"
    os.chmod(path, stat.S_IXUSR | os.stat(path).st_mode)


def is_executable(path):
    "check if the file is executable by the current user (u+x)"
    return os.stat(path)[0] & stat.S_IXUSR


def real_basename(x):
    """ a 'better' basename (removes the trailing / like on Unix) """
    x = x.rstrip(os.sep)
    return os.path.basename(x)


def multi_glob(pats, exclude=None):
    """ glob using a list of patterns and removing duplicate files, exclude name in the list for which the callback exclude(name) return true
    example: advanced_glob(['*.jpg','*.gif'],exclude=lambda n:len(n)>20) return a list of all JPG and GIF files which have names shorter then 20 characters
    """

    unique = {}
    if exclude is None:
        def exclude(n): return 0

    for p in pats:
        for f in glob.glob(p):
            unique[f] = 1

    return [name for name in unique.keys() if not exclude(name)]


def recursive_copy(src, dest):
    """ copy src file (or a directory tree if src specifies a directory) to dest directory. dest must be a directory and must exist.
    if src is a relative path, then the src directory structure is preserved in dest.
    """
    
    if not os.path.isdir(dest):
        raise ValueError(
            'resursive_copy: destination %s must specify a directory (which exists)' % dest)

    if os.path.isdir(src):
        destdir = dest
        srcdir, srcbase = os.path.split(src.rstrip('/'))
        if not srcdir == '' and not os.path.isabs(src):
            destdir = os.path.join(destdir, srcdir)
            if not os.path.isdir(destdir):
                os.makedirs(destdir)
        shutil.copytree(src, os.path.join(destdir, srcbase))
    else:

        srcdir = os.path.dirname(src.rstrip('/'))
        if srcdir == '' or os.path.isabs(src):
            shutil.copy(src, dest)
        else:
            destdir = os.path.join(dest, srcdir)
            if not os.path.isdir(destdir):
                os.makedirs(destdir)
            shutil.copy(src, destdir)


def remove_prefix(fn, path_list):
    """Remove the common prefix of fn and the first matching element in the path_list.

    Example: for fn='/a/b/c' and path=['d','/a','/a/b'] return 'b/c'

    If no matching path is found, then fn is returned unchanged.

    This function converts each element of the path_list using realpath.abspath.

    """
    for p in path_list:
        # normalize path
        if not p or p == '.':
            continue
        p = os.path.realpath(os.path.abspath(p))
        # print 'searching path component: %s'%p
        idx = fn.find(p)
        if idx != -1:
            # print 'found "%s" atr index %d in "%s"'%(p,idx,fn)
            return fn[len(p) + len(os.sep):]

    return fn

if __name__ == "__main__":

    workdir = 'test_recursive_copy'
    shutil.rmtree(workdir, True)
    os.mkdir(workdir)
    os.chdir(workdir)

    destdir = 'dest'
    os.mkdir(destdir)

    # Copy file
    os.mknod('file1')
    recursive_copy('file1', destdir)
    assert os.path.isfile(destdir + '/file1')

    # Copy file in dir
    os.mkdir('src')
    os.mknod('src/file2')
    recursive_copy('src/file2', destdir)
    assert os.path.isfile(destdir + '/src/file2')

    # Copy file in dir
    #       Check possibility to copy to existing dir
    os.mknod('src/file3')
    recursive_copy('src/file3', destdir)
    assert os.path.isfile(destdir + '/src/file3')

    # Copy file with abs path
    os.mknod('file4')
    recursive_copy(os.getcwd() + '/file4', destdir)
    assert os.path.isfile(destdir + '/file4')

    # Copy subdir
    os.mkdir('src/subdir')
    os.mknod('src/subdir/file5')
    recursive_copy('src/subdir', destdir)
    assert os.path.isfile(destdir + '/src/subdir/file5')

    # Copy subdir
    #       Check "/" at the end
    os.mkdir('src/subdir2')
    os.mknod('src/subdir2/file6')
    recursive_copy('src/subdir2/', destdir)
    assert os.path.isfile(destdir + '/src/subdir2/file6')

    # Copy subdir with abs path
    os.mkdir('src/subdir3')
    os.mknod('src/subdir3/file7')
    recursive_copy(os.getcwd() + '/src/subdir3', destdir)
    assert os.path.isfile(destdir + '/subdir3/file7')



###############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: ArgSplitter.py,v 1.1 2008-07-17 16:40:59 moscicki Exp $
###############################################################################

import copy
from GangaCore.GPIDev.Adapters.ISplitter import ISplitter
from GangaCore.GPIDev.Base.Proxy import addProxy, stripProxy
from GangaCore.GPIDev.Schema import Schema, Version, SimpleItem
from GangaCore.GPIDev.Lib.Dataset.GangaDataset import GangaDataset
from GangaLSST.Lib.Im3ShapeApp.Im3ShapeApp import Im3ShapeApp

from GangaCore.Utility.logging import getLogger
logger = getLogger()


class Im3ShapeSplitter(ISplitter):

    """
    This splitter splits jobs using the Im3ShapeApp application using the size parameter.

    If a splitter is configured with size = 5, split_by_file = True, then it will create 5 subjobs per file in the master_job.inputdata
    If a splitter is configured with size = 5, split_by_file = False, then it will create 5 subjobs total and configure all subjobs to use all given data.

    In the future there may be support for splitting based upon regex and namePatterns in the inputdata to allow a certain subset of data to be put in each subjob.
    """
    _name = "Im3ShapeSplitter"
    _schema = Schema(Version(1, 0), {'size': SimpleItem(defvalue=5, doc='Size of the tiles which are to be split.'),
                                     'split_by_file': SimpleItem(defvalue=True, doc='Should we auto-split into subjobs here on a per-file basis?')})

    def split(self, job):
        """
        Actually perform the splitting of the given master job. The generated subjobs of the splitting are returned
        Args:
            job (Job): This is the master job object which is to be split and the subjobs of which are returned
        """

        assert isinstance(job.application, Im3ShapeApp)

        subjobs = []

        def getApp(job, rank, size):
            app = copy.deepcopy(job.application)
            app.rank = rank
            app.size = size
            return app

        if self.split_by_file:
            for this_file in job.inputdata:
                for rank in range(0, self.size):
                    j = self.createSubjob(job, ['application'])
                    # Add new arguments to subjob
                    j.application = getApp(job, rank, self.size)
                    j.inputdata = GangaDataset(files = [stripProxy(this_file)])
                    subjobs.append(j)
        else:
            for rank in range(0,self.size):
                j = self.createSubjob(job, ['application'])
                j.application = getApp(job, rank, self.size)
                j.inputdata = job.inputdata
                logger.debug('Rank for split job is: ' + str(rank))
                subjobs.append(j)

        return subjobs




from setuptools import setup
from setuptools import find_packages
from setuptools import Command
import subprocess
import os
import sys

file_path = os.path.dirname(os.path.realpath(__file__))

_gangaVersion = '8.3.1'

def version():
    return _gangaVersion


def readme():
    filename = os.path.abspath(os.path.join(file_path, 'README.rst'))
    with open(filename) as f:
        return f.read()


class RunTestsCommand(Command):
    """
    A custom setuptools command to run the Ganga test suite.
    """
    description = 'run the Ganga test suite'
    all_types = ['unit', 'integration', 'all']
    user_options = [
        ('type=', 't', 'the type of tests: [{0}]'.format(', '.join(all_types))),
        ('coverage', None, 'should coverage be generated'),
        ('xunit', None, 'should xunit-compatible files be produced'),
    ]

    def initialize_options(self):
        self.type = 'unit'
        self.coverage = False
        self.xunit = False

    def finalize_options(self):
        if self.type not in self.all_types:
            raise Exception('Test type must be [{0}]'.format(', '.join(self.all_types)))

    @staticmethod
    def _get_test_env():
        ganga_python_dir = os.path.join(file_path, 'ganga')

        test_env = os.environ.copy()
        path = ':'.join(s for s in [ganga_python_dir, test_env.get('PYTHONPATH', None)] if s)
        test_env['PYTHONPATH'] = path

        return test_env

    def run(self):

        cmd = ['py.test']

        if self.type in ['unit', 'all']:
            cmd.append('ganga/GangaCore/test/Unit')
            cmd.append('ganga/GangaCore/Core')
            cmd.append('ganga/GangaCore/Runtime')
            cmd.append('ganga/GangaCore/Utility')
        if self.type in ['integration', 'all']:
            cmd.append('ganga/GangaCore/test/GPI')

        if self.coverage:
            cmd.append('--cov-report xml --cov .')
        if self.xunit:
            cmd.append('--junitxml tests.xml')

        subprocess.check_call(' '.join(cmd), cwd=file_path, shell=True, env=self._get_test_env())


pythonPackages = find_packages('./')
pythonPackages.append('ganga/GangaRelease')

setup(name='ganga',
      description='Job management tool',
      long_description=readme(),
      url='https://github.com/ganga-devs/ganga',
      version=version(),
      author='Ganga Developers',
      author_email='project-ganga-developers@cern.ch',
      license='GPL v2',
      scripts=['bin/ganga'],
      package_dir={'ganga':'ganga', 'GangaRelease':'ganga/GangaRelease'},
      packages=pythonPackages,
      install_requires=[
          'ipython>=5.0.0',
          'httplib2>=0.8',
          'absl-py>=0.1.2',
          'google-api-python-client',
          'google-auth-httplib2',
          'google-auth-oauthlib',
          'requests>=2.23.0',
          'Flask~=1.1.2'
      ],
      extras_require={'profiler' : ['memory_profiler'], 'LHCb' : ['LbDevTools']},
      classifiers=[
          'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
          'Programming Language :: Python :: 3.6',
      ],
      include_package_data=True,
      package_data={'GangaCore': ['Runtime/HEAD_CONFIG.INI'], 'GangaRelease':['ReleaseNotes-*', 'tools/check-new-ganga.py', 'tools/ganga-cvmfs-install.sh', 'tools/ganga-cvmfs-install-dev.sh']},
      cmdclass={
          'tests': RunTestsCommand,
      },
      )




from GangaCore.Core.GangaRepository import addRegistry

from .PrepRegistry import PrepRegistry
from .JobRegistry import JobRegistry
from .BoxRegistry import BoxRegistry

addRegistry(PrepRegistry("prep", "stuff"))
addRegistry(JobRegistry("jobs", "General Job Registry"))
addRegistry(JobRegistry("templates", "Templates"))
addRegistry(BoxRegistry("box", "The Ganga box"))



# File: GangaTutorial/__init__.py

def loadPlugins( config = {} ):

    from .Lib import PrimeFactorizer
    from .Lib import PrimeFactorizerSplitter
    from .Lib import PrimeTableDataset
	
    return None
